"""
Copyright (c) 2017, 2018, 2019 Irstea
Copyright (c) 2017, 2018, 2019 François Kneib
Copyright (c) 2017, 2018, 2019 Franck Bourrier
Copyright (c) 2017, 2018, 2019 David Toe
Copyright (c) 2017, 2018, 2019 Frédéric Berger
Copyright (c) 2017, 2018, 2019 Stéphane Lambert

This file is part of PlatRock.

PlatRock is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, version 3 of the License.

PlatRock is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Foobar. If not, see <http://www.gnu.org/licenses/>.
"""

"""
This module handles all the engines (the core) of the ThreeD code.
TODO : add rock-terrain and rock-tree Cython functions.
"""

from . import ThreeDEnginesToolbox
import platrock.Common.Debug as Debug
import platrock.Common.Math as Math
import platrock.Common.Outputs as Out
import numpy as np
from . import Objects
import quaternion, shapely
import copy
import types

class Engine(object):
	"""
	"Abstract" class for all engines. Each engine has a run() method which is run at each "iter_every" iteration. Engines instances must be stored in :attr:`ThreeD.Simulations.Simulation.Engines`.
	
	Args:
		dead (bool): if True, the engine will not be ran
		iter_every (int): run the engine each iter_every timesteps
	"""
	def __init__(self,iter_every=1,use_cython=False):
		self.dead=False
		self.iter_every=iter_every
		self.use_cython=use_cython


class Verlet_update(Engine):
	"""
	Update the rocks faces (:attr:`ThreeD.Objects.Rock.verlet_faces_list`) and trees (:attr:`ThreeD.Objects.Rock.verlet_trees_list`) neighbours lists.
	
	Args:
		dist_factor (float): the verlet distance factor (dist_factor * rock_radius = verlet distance). Values must be >1, but values >5 are highly recommanded.If dist_factor==1, the verlet algorithm will be unefficient.
		
	"""
	def __init__(self,dist_factor=5,**kwargs):
		super(Verlet_update,self).__init__(**kwargs)
		self.dist_factor=dist_factor
		if(self.use_cython):
			self.run=types.MethodType(ThreeDEnginesToolbox.verlet_run,self)
		else:
			def run(self,s):
				r=s.current_rock
				if(not((r.pos-r.radius*1.1<r.bounding_box.pos-r.bounding_box.half_length).any() or (r.pos+r.radius*1.1>r.bounding_box.pos+r.bounding_box.half_length).any())): #if the rock stays inside the box
					return
				r.bounding_box.half_length=self.dist_factor*r.radius
				r.bounding_box.pos[:]=r.pos[:]
				
				# ROCK - TERRAIN verlet list :
				r.verlet_faces_list=[]	#initialize the list
				mins_check_outside = (r.bounding_box.pos+r.bounding_box.half_length<s.terrain.faces_xyz_bb[:,0::2]).any(axis=1)
				maxs_check_outside = (r.bounding_box.pos-r.bounding_box.half_length>s.terrain.faces_xyz_bb[:,1::2]).any(axis=1)
				inside_mask = np.logical_not(np.logical_or(mins_check_outside, maxs_check_outside))
				r.verlet_faces_list=s.terrain.faces[inside_mask].tolist()
				
				if(s.enable_forest):
					r.verlet_trees_list=[]
					# ROCK - TREES verlet list :
					dists=((s.terrain.trees_as_array[:,0:2] - r.pos[0:2])**2).sum(axis=1)-(s.terrain.trees_as_array[:,2]/2)**2# all (rock centroid) - (trees) squared distances. NOTE: trees_as_array[:,2] are the dhp per trees
					active_indices=np.where(dists<(r.bounding_box.half_length)**2)[0]	#indices of trees composing the verlet list
					r.verlet_trees_list=np.asarray(s.terrain.trees)[active_indices]
					#Debug.info("len(verlet_trees_list) =",len(r.verlet_trees_list))
				
			self.run=types.MethodType(run,self)

class Contacts_detector(Engine):
	"""
	FIXME:DOC
	Append (real, geometrical) contacts into :attr:`ThreeD.Objects.Rock.terrain_active_contacts` dict. Rock-Face goes into :attr:`terrain_active_contacts["terrain"]<ThreeD.Objects.Rock.>`  and Rock-Tree contacts goes into :attr:`terrain_active_contacts["tree"]<ThreeD.Objects.Rock.terrain_active_contacts>` .
	"""
	def __init__(self,**kwargs):
		super(Contacts_detector,self).__init__(**kwargs)
		if(self.use_cython):
			self.run=types.MethodType(ThreeDEnginesToolbox.contacts_detector_run,self)
		else:
			def run(self,s):
				
				# ROCK-TERRAIN :
				r=s.current_rock
				r.terrain_active_contact=None
				face_candidate=None
				face_candidate_dist=np.inf
				#for rp in r.points:
				rp=r.points[0] #for a Object.Sphere there is a unique point
				for tf in r.verlet_faces_list:
					dist=tf.normal.dot(rp.pos-tf.points[0].pos)-r.radius	# first pass : point-plane distance. Negative dist <=> interpenetration.
					if(dist>0.): # no penetration with the plan : no contact possible
						continue
					if(r.vel.dot(tf.normal)>0): #if the rock goes appart from the face. This check was added with the edge and wedge contact detection: edge and wedge detection algorithms could lead to select a face the rock is going appart from.
						continue
					# We now search the nearest distance between the sphere and the triangle. 3 possibilities: surface contact, edge contact, vertex contact
					if(tf.is_point_inside_3D(rp)):
						#we are sure that the contact is on the face surface
						if(dist<face_candidate_dist):
							#this is currently the best candidate
							face_candidate=tf
							face_candidate_dist=dist
						continue 	#as the contact is on the face surface, it is not on an edge or on a vertex. So continue to the next face
					# We now search an edge contact
					found_edge_contact=False
					for i in range(-1,2):
						branch=rp.pos-tf.points[i].pos
						edge=tf.points[i+1].pos-tf.points[i].pos
						proj_coef=branch.dot(edge)/(branch.norm()**2)
						if(proj_coef>1 or proj_coef<0):continue #the projection of the sphere center on the edge is outside the edge
						proj=edge*proj_coef
						dist=(branch-proj).norm()-r.radius
						if(dist<face_candidate_dist and dist<0):
							#this is currently the best candidate
							face_candidate=tf
							face_candidate_dist=dist
							found_edge_contact=True
					if(found_edge_contact):continue	#as the contact is on an edge, it is not on a vertex. So continue to the next face
					# We now search a vertex contact
					for i in range(0,3):
						dist=(tf.points[i].pos-rp.pos).norm()-r.radius
						if(dist<face_candidate_dist and dist<0):
							#this is currently the best candidate
							face_candidate=tf
							face_candidate_dist=dist
				if(face_candidate_dist<0):	#the default value for face_candidate_dist is +infinity, see above.
					r.terrain_active_contact=Objects.Rock_terrain_contact(rp,face_candidate)
					r.terrain_active_contact.dist=face_candidate_dist
				
				# ROCK - TREES :
				if(s.enable_forest):
					r.tree_active_contact=None
					for t in r.verlet_trees_list:
						if(t.active):
							dist=Math.norm2(r.pos[0:2]-t.pos)-r.radius-t.dhp/2./100	#negative dist <=> interpenetration
							if(dist<0.):
								r.tree_active_contact=Objects.Rock_tree_contact(t)
								r.tree_active_contact.dist=dist
			self.run=types.MethodType(run,self)
		
		

class Nscd_integrator(Engine):
	"""
	FIXME : doc
	"""
	def __init__(self,**kwargs):
		super(Nscd_integrator,self).__init__(**kwargs)
		if(self.use_cython):
			self.run=types.MethodType(ThreeDEnginesToolbox.nscd_integrator_run,self)
		else:
			def run(self,s):
				r=s.current_rock
				r.vel[2]-=s.gravity*s.dt
				r.pos+=r.vel*s.dt
				if(s.GUI_enabled):
					#/!\ NOTE this rotational integration scheme is only valid for spheres
					r.ori=quaternion.from_rotation_vector(r.angVel*s.dt)*r.ori
					r.ori=r.ori.normalized()
				r.update_members_pos()
				if(r.pos[2]<s.terrain.min_height):
						Debug.warning("The rock went outside the terrain !")
						r.out_of_bounds=True
						r.is_stopped=True
			self.run=types.MethodType(run,self)

class Rock_terrain_nscd_basic_contact(Engine):
	"""
	FIXME : doc
	"""
	def __init__(self,**kwargs):
		super(Rock_terrain_nscd_basic_contact,self).__init__(**kwargs)
	def run(self,s):
		r=s.current_rock
		if(r.terrain_active_contact):
			f=r.terrain_active_contact.face
			if(s.override_rebound_params):
				bounce_model_number=s.override_bounce_model_number
			else:
				bounce_model_number=f.bounce_model_number
			BM=s.number_to_model_correspondance[bounce_model_number]
			BM.run(r,f)
			cp=r.pos-r.radius*f.normal
			#Push the sphere outside the face along the z axis. Use the z axis and not face.normal as the latter would cause bugs in Posprocessings.
			r.pos[2]+=-(f.normal[0]*(cp[0]-f.points[0].pos[0]) + f.normal[1]*(cp[1]-f.points[0].pos[1]))/f.normal[2]+f.points[0].pos[2]-cp[2]+0.01*r.radius
			s.output.add_contact(r,BM.updated_normal,Out.SOIL)

class Rock_tree_nscd_basic_contact(Engine):
	"""
	FIXME : doc
	"""
	def __init__(self,**kwargs):
		super(Rock_tree_nscd_basic_contact,self).__init__(**kwargs)
	def run(self,s):
		r=s.current_rock
		if(not r.terrain_active_contact and r.tree_active_contact):
			t=r.tree_active_contact.tree
			Debug.info("Rock-tree contact !")
			vel=r.vel.norm()
			vol=r.volume
			
			vxy_rock=Math.Vector2(r.vel[0:2])
			xy_contact_point=t.pos+t.dhp/2/100*Math.normalized2(r.pos[0:2]-t.pos)
			ecc_axis=Math.Vector2([-vxy_rock[1],vxy_rock[0]]).normalized() #the normed axis, orthogonal to vxy, on which the eccentricity is measured from the contact point and vxy
			ecc=(xy_contact_point-t.pos).dot(ecc_axis)/(t.dhp/2./100)	#eccentricity, from -1 to 1
			ang=np.sign(r.vel[2])*abs(np.degrees(np.arctan(r.vel[2]/vxy_rock.norm()))) # the vertical incident angle.
			angle=np.arctan2(vxy_rock[1],vxy_rock[0])	#the signed angle between the rock xy velocity and the +x axis. NOTE: arctan2=arctan2(y,x)
			input_rock_vel_norm=r.vel.norm()
			r.vel[:]=s.forest_impact_model.get_output_vel(vel,vol,abs(ecc),t.dhp,ang) #we use abs(ecc) as the DEM model is symetric, see the line below for ecc<0. The vel here is expressed in Toe's C.S.
			if(ecc<0):r.vel[1]*=-1
			r.vel[0:2]=Math.rotated2(r.vel[0:2],angle) #rotate the vel to translate from Toe's C.S to our global C.S.
			r.vel=r.vel.normalized()*input_rock_vel_norm*r.vel.norm()/s.forest_impact_model.last_computed_data["input_vel"] #linear interpolation of velocity norm between PlatRock input vel and Toe's input one.
			normal=r.pos.copy()
			normal[0:2]-=t.pos
			normal=normal.normalized()
			s.output.add_contact(r,normal,Out.TREE)
			t.active=False
			if(s.GUI_enabled):
				t.color=[0.8,0,0]
				s.GUI.updateTreeColor(t)
			
			
class Snapshooter(Engine):
	def __init__(self,filename="snap_",**kwargs):
		super(Snapshooter,self).__init__(**kwargs)
		self.filename=filename
		self.ndone=0
	def run(self,s):
		s.GUI.take_snapshot(self.filename+str(self.ndone).rjust(6,"0")+".png")
		self.ndone+=1
		
class Time_stepper(Engine):
	def __init__(self,safety_coefficient=0.1,**kwargs):
		super(Time_stepper,self).__init__(**kwargs)
		self.safety_coefficient=safety_coefficient
	def run(self,s):
		s.dt=self.safety_coefficient*s.current_rock.radius/(s.current_rock.vel.norm())
		s.dt=min(s.dt,0.1)
		#Debug.warning("New time step:",s.dt)
		


class Siconos(Engine):
	def __init__(self,terrain=None,dt=0.02,**kwargs):
		super(Siconos,self).__init__(**kwargs)
		from siconos.kernel import NonSmoothDynamicalSystem, MoreauJeanOSI, TimeDiscretisation, TimeStepping, RollingFrictionContact, SICONOS_TS_NONLINEAR, NewtonImpactRollingFrictionNSL,MLCPProjectOnConstraints
		from siconos.numerics import SICONOS_FRICTION_3D_NSGS, SICONOS_FRICTION_3D_NSGS_ERROR_EVALUATION_LIGHT, SICONOS_FRICTION_3D_NSGS_FILTER_LOCAL_SOLUTION_TRUE, SICONOS_FRICTION_3D_ONECONTACT_NSN_GP_HYBRID,SICONOS_MLCP_ENUM
		from siconos.mechanics.collision.bullet import SiconosBulletCollisionManager, SiconosBulletOptions
		from siconos.io.io_base import MechanicsIO
		
		self.io = MechanicsIO()
		self.nsds = NonSmoothDynamicalSystem(0,np.inf)
		self.timedisc=TimeDiscretisation(0, dt)#FIXME FIXME FIXME : set DT here !
		
		self.osi=MoreauJeanOSI(0.50001)
		
		self.osnspb=RollingFrictionContact(5)#,SICONOS_FRICTION_3D_NSGS)
		self.osnspb.setMaxSize(30000)	#NOTE: what is that ?
		#self.osnspb.setMStorageType(1)
		self.osnspb.setNumericsVerboseMode(False)
		self.osnspb.setKeepLambdaAndYState(True)
		solverOptions = self.osnspb.numericsSolverOptions()
		solverOptions.iparam[0] = 100		#iter max
		solverOptions.dparam[0] = 1e-3			#tolerance
		solverOptions.iparam[1] = SICONOS_FRICTION_3D_NSGS_ERROR_EVALUATION_LIGHT
		solverOptions.iparam[14] = SICONOS_FRICTION_3D_NSGS_FILTER_LOCAL_SOLUTION_TRUE
		#fcOptions = solverOptions.internalSolvers[0]
		#fcOptions.iparam[0] = 20
		#fcOptions.solverId = SICONOS_FRICTION_3D_ONECONTACT_NSN_GP_HYBRID
		
		#self.osnspb_pos=MLCPProjectOnConstraints(SICONOS_MLCP_ENUM, 1.0)
		#self.osnspb_pos.setMaxSize(30000)
		##self.osnspb_pos.setMStorageType(0) # "not yet implemented for sparse storage"
		#self.osnspb_pos.setNumericsVerboseMode(False)
		#self.osnspb_pos.setKeepLambdaAndYState(True)
		
		self.sim=TimeStepping(self.nsds,self.timedisc)#,self.osi,self.osnspb,self.osnspb_pos)
		#self.sim.setProjectionMaxIteration(20)
		#self.sim.setConstraintTolUnilateral(1e-5);
		#self.sim.setConstraintTol(1e-5);
		self.sim.insertIntegrator(self.osi)
		self.sim.insertNonSmoothProblem(self.osnspb)
		
		self.interman=SiconosBulletCollisionManager(SiconosBulletOptions())
		self.sim.insertInteractionManager(self.interman)
		
		self.sim.setNewtonOptions(SICONOS_TS_NONLINEAR)
		self.sim.setNewtonMaxIteration(1)
		self.sim.setNewtonTolerance(1e-3)
		self.add_terrain(terrain)
		self.nsl=NewtonImpactRollingFrictionNSL(0.,0.,0.,0.6,5) #NOTE: en, et, mu, ?
		self.interman.insertNonSmoothLaw(self.nsl, 1, 2)
		
	def add_terrain(self,terrain):
		from siconos.mechanics.collision import SiconosContactor, SiconosContactorSet
		cset = SiconosContactorSet()
		csetpos = [0,0,0,1,0,0,0]
		hm,offsets=terrain.get_siconos_heightmap()
		cset.append(SiconosContactor(hm, [offsets[0],offsets[1],0,1,0,0,0], 1))
		#if(terrain.enable_forest):	#FIXME: enable the forest flag
		from siconos.mechanics.collision import SiconosCylinder
		for t in terrain.trees:
			lowest_terrain_point_z=terrain.points_as_array[:,2].min()
			highest_terrain_point_z=terrain.points_as_array[:,2].max()+50.
			cylinders_height=highest_terrain_point_z-lowest_terrain_point_z
			cyl=SiconosCylinder(t.dhp/2/100,cylinders_height)
			q=quaternion.from_rotation_vector([np.pi/2,0,0])
			cset.append(SiconosContactor(cyl, [t.pos[0],t.pos[1],lowest_terrain_point_z+cylinders_height/2.,q.w,q.x,q.y,q.z], 1))
		self.interman.insertStaticContactorSet(cset, csetpos)
	
	def add_rock(self,r):
		from siconos.mechanics.collision import RigidBodyDS, SiconosContactorSet, SiconosContactor
		self.DS = RigidBodyDS(r.pos.tolist() + [r.ori.w,r.ori.x,r.ori.y,r.ori.z], r.vel.tolist(), r.mass, 0.0*np.eye(3))#NOTE:  0.0*np.eye(3) is the inertia
		cset = SiconosContactorSet()
		csetpos = [0,0,0,1,0,0,0]
		cset.append(SiconosContactor(r.get_siconos_convex_hull(), csetpos, 2))
		self.DS.setContactors(cset)
		#self.DS.setNumber(np.random.randint(1e10)) #NOTE : set something else maybe
		self.nsds.insertDynamicalSystem(self.DS)
		#self.nsds.setName(self.DS, str(name))	#NOTE: set a name maybe
		self.DS.setFExtPtr([0, 0, - self.DS.scalarMass() * 9.8])#FIXME: set s.g intead of hardcoding

	def remove_rock(self):
		self.interman.removeBody(self.DS)
		self.nsds.removeDynamicalSystem(self.DS)
	
		
	
	def run(self,s):
		ds_nb=self.nsds.getNumberOfDS()-1
		shapely_point=shapely.geometry.Point(s.current_rock.pos[0:2])
		for i in range(len(s.terrain.soil_params)):	
			if(s.terrain.soil_params[i]["shapely_polygon"].contains(shapely_point)): #NOTE: this could be optimized by avoiding updating mu and e when the rock doesn't change from polygon (when i==0)
				e=s.terrain.soil_params[i]["params"].get("e",s.terrain.default_faces_params["e"])
				mu=s.terrain.soil_params[i]["params"].get("mu",s.terrain.default_faces_params["mu"])
				mu_r=s.terrain.soil_params[i]["params"].get("mu_r",s.terrain.default_faces_params["mu_r"])
				self.nsl.setEn(e)
				self.nsl.setMu(mu)
				self.nsl.setMuR(mu_r)
				s.terrain.soil_params.insert(0,s.terrain.soil_params.pop(i))	#swap poly_params sequence: put the current one at the first position to increase speed
				break
			#if no polygon found:
			e=s.terrain.default_faces_params["e"]
			mu=s.terrain.default_faces_params["mu"]
			mu_r=s.terrain.default_faces_params["mu_r"]
		self.sim.computeOneStep()
		s.current_rock.pos[:]=self.DS.q()[0:3]
		s.current_rock.vel[:]=self.DS.velocity()[0:3]
		s.current_rock.angVel[:]=self.DS.velocity()[3:]
		s.current_rock.ori.w=self.DS.q()[3]
		s.current_rock.ori.x=self.DS.q()[4]
		s.current_rock.ori.y=self.DS.q()[5]
		s.current_rock.ori.z=self.DS.q()[6]
		if(self.nsds.getNumberOfInteractions()>0 and (self.io.contactPoints(self.nsds) is not None)):
			normal=Math.normalized3(self.io.contactPoints(self.nsds)[:,7:10].mean(axis=0))
			s.output.add_contact(s.current_rock,normal,Out.SOIL)
		if(s.GUI_enabled):
			s.current_rock.update_members_pos()
		if(s.current_rock.pos[2]<s.terrain.min_height):
			Debug.warning("The rock went outside the terrain !")
			s.current_rock.out_of_bounds=True
			s.current_rock.is_stopped=True
		self.sim.clearNSDSChangeLog()
		self.sim.nextStep()





#...........::::::::::::: DEPRECATED/UNDER DEVELOPMENT :::::::::::............



class Newton_integrator(Engine):
	"""
	Convert the force and torque applied to the rock into translation and rotation. Acutally move/rotate the rock. This engine also stores copies of the contacts into :attr:`ThreeD.Objects.Rock.contacts` list.
	"""
	def run(self,s):
		r=s.current_rock
		
		# TRANSLATION :
		linAccel=r.force/r.mass + np.array([0.,0.,-s.gravity])
		r.vel+=s.dt*linAccel
		r.pos+=r.vel*s.dt
		
		# ROTATION :
		A				=	quaternion.as_rotation_matrix(r.ori.conjugate())			# rotation matrix from global to local r.f.
		l_n				=	r.angMom + s.dt/2. * r.torque								# global angular momentum at time n
		l_b_n			=	np.dot(A,l_n)												# local angular momentum at time n
		angVel_b_n		=	l_b_n/r.I													# local angular velocity at time n
		dotQ_n			=	Math.DotQ(angVel_b_n,r.ori)									# dQ/dt at time n
		Q_half			=	r.ori + s.dt/2. * dotQ_n									# Q at time n+1/2
		r.angMom		+=	s.dt*r.torque												# global angular momentum at time n+1/2
		l_b_half		=	np.dot(A,r.angMom)											# local angular momentum at time n+1/2
		angVel_b_half	=	l_b_half/r.I												# local angular velocity at time n+1/2
		dotQ_half		=	Math.DotQ(angVel_b_half,Q_half)								# dQ/dt at time n+1/2
		r.ori			=	r.ori + s.dt*dotQ_half										# Q at time n+1
		r.angVel		=	quaternion.rotate_vector(r.ori,angVel_b_half)				# global angular velocity at time n+1/2
		r.ori			=	r.ori.normalized()
		
		
		# MOVE THE POINTS ACCORDINGLY
		r.update_members_pos()
		
		# EVENTUALLY STORE THE CONTACT PROPERTIES FOR POST-PROCESSING
		if(len(r.terrain_active_contacts["terrain"])>0): #NOTE : r.terrain_active_contacts["terrain"] normally contains only one element. NOTE 2 : make hard copy of all variables
			r.contacts.append(r.terrain_active_contacts["terrain"][0].get_storage_copy(r))
		#FIXME : uncomment below to store rock-tree contacts when postprocessing will allow it
		#if(len(r.terrain_active_contacts["tree"])>0):
			#r.contacts.append(r.terrain_active_contacts["tree"][0].get_storage_copy(r))
			
		
class Rock_terrain_nscd_contact(Engine):
	"""
	Choose the contact for which the interpenetration is maximal and append the corresponding force/torque on the rock. NSCD stands for non-smooth contact-dynamics, meaning that the contact is instantaneous: from the input velocity and the contact properties, we compute the output velocity. Then the force/torque corresponding to the (input-output) velocity difference are computed.
	"""
	def run(self,s):
		maxDist=np.inf
		contact=None
		r=s.current_rock
		for c in r.terrain_active_contacts["terrain"]:	#only keep the most interpenetrated contact
			if(c.status==1 and maxDist>c.dist):
				contact=c
				maxDist=c.dist
		if(contact):
			r.terrain_active_contacts["terrain"]=[contact] #remove multiple contacts here
			p=contact.point
			f=contact.face
			
			#Compute point velocity and decompose
			p.vel=r.vel+(r.pos-p.pos).cross(r.angVel)
			p_vel_n=p.vel.dot(f.normal)*f.normal
			p_vel_t=p.vel-p_vel_n
			
			#Apply rebound
			p_vel_n *= -r.R_n
			p_vel_t *= r.R_t
			
			#Compute velocity difference
			delta_p_vel = (p_vel_n+p_vel_t) - p.vel
			
			#Compute the force/torque needed to get the velocity difference

			r.force  +=  r.mass/s.dt*delta_p_vel
			r.torque +=  (p.pos-r.pos).cross(r.force)
			
			#Position correction to avoid real penetration of the rock into the face
			r.pos-=f.normal*contact.dist*1.001
			
			#The contact is computed, set the status accordingly
			contact.status=2
			
class Rock_tree_basic_contact(Engine):
	"""
	Choose the contact for which the interpenetration is maximal and append the corresponding force/torque on the rock. The contact is instantaneous: from the input velocity and the contact properties, we compute the output velocity. Then the force/torque corresponding to the (input-output) velocity difference are computed.
	"""
	def run(self,s):
		maxDist=np.inf
		contact=None
		r=s.current_rock
		for c in r.terrain_active_contacts["tree"]:	#only keep the most interpenetrated contact
			if(c.status==1 and maxDist>c.dist):
				contact=c
				maxDist=c.dist
		if(contact):
			r.terrain_active_contacts["tree"]=[contact] #remove multiple contacts here
			normal=(r.pos[0:2]-contact.tree.pos).normalized()
			contact.pos=Math.Vector3([0.,0.,r.pos[2]])
			contact.pos[0:2]=contact.tree.pos+(contact.tree.dhp/2./100)*normal
			
			normalized_rock_vel=r.vel[0:2].normalized()
			contact.eccentricity=( contact.pos[0:2] + (contact.tree.pos - contact.pos[0:2]).dot(normalized_rock_vel)*normalized_rock_vel - contact.tree.pos ).norm() / (contact.tree.dhp/2./100)
			
			#Compute the force needed to get the velocity difference
			v_factor=0.8+0.2*contact.eccentricity
			vel_n=r.vel[0:2].dot(normal)*normal
			vel_t=r.vel[0:2]-vel_n
			new_vel_2D = -vel_n + vel_t
			delta_v = Math.Vector3([new_vel_2D[0]*v_factor,new_vel_2D[1]*v_factor,r.vel[2]]) - r.vel
			r.force+=r.mass/s.dt*delta_v
			
			#Position correction to avoid real penetration of the rock into the tree
			r.pos[0:2] -= normal*contact.dist*1.001
			
			contact.tree.active=False
			contact.tree.color=[0.8,0,0]
			if(s.GUI_enabled):
				s.GUI.updateTreeColor(contact.tree)
			contact.status=2
		
		
		
		
		
		
		
		



"""
Copyright (c) 2017, 2018, 2019 Irstea
Copyright (c) 2017, 2018, 2019 François Kneib
Copyright (c) 2017, 2018, 2019 Franck Bourrier
Copyright (c) 2017, 2018, 2019 David Toe
Copyright (c) 2017, 2018, 2019 Frédéric Berger
Copyright (c) 2017, 2018, 2019 Stéphane Lambert

This file is part of PlatRock.

PlatRock is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, version 3 of the License.

PlatRock is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Foobar. If not, see <http://www.gnu.org/licenses/>.
"""

"""
This module is used by the TwoD model. It describes 2D geometric objects and provide some functions to solve geometric problems.
"""

import numpy as np
import platrock.Common.Debug as Debug
import platrock.Common.Math as Math

class Parabola(object):
	"""
	A 2D parabola, described by : :math:`z = Ax^2 + Bx +C`
	
	Args:
		A (float): the :math:`A` square parameter
		B (float): the :math:`B` linear parameter
		C (float): the :math:`C` constant parameter
	"""
	def __init__(self,r=None,pos=None,vel=None,g=None,ABC=None): #r is a rock
		"""
		Constructor, {r,g} or {pos,vel,g}
		
		Args:
			r (:class:`TwoD.Object.Rock`): deduce the parabola from a rock
			pos (:class:`Common.Math.Vector2` [x,z]): deduce the parabola from a position and a velocity
			vel (:class:`Common.Math.Vector2` [vx,vz]): deduce the parabola from a position and a velocity
			g (float): gravity acceleration
		"""
		if(not r==None):
			x=r.pos[0] ; y=r.pos[1]
			vx=r.vel[0] ; vy=r.vel[1]
		elif(isinstance(pos,Math.Vector2) and isinstance(vel,Math.Vector2)):
			x=pos[0] ; y=pos[1]
			vx=vel[0] ; vy=vel[1]
		elif(ABC is not None):
			self.A=ABC[0]
			self.B=ABC[1]
			self.C=ABC[2]
			return
		self.A=-g/(2.*vx**2)
		self.B=vy/vx+g*x/vx**2
		self.C=y-g*x**2/(2.*vx**2)-vy*x/vx
	
	def get_value_at(self,x):
		"""
		Returns the :math:`z` value of the parabola at the given :math:`x` position.
		
		Args:
			x (float): the :math:`x` value
			
		Returns:
			float: :math:`Ax^2 + Bx +C`
		"""
		return self.A*x*x + self.B*x + self.C
	
	def get_arrow_x_from_gradient(self,gradient):
		"""
		Returns the :math:`x` value of the arrow for this parabola and a slope defined by its gradient. If the parabola is a trajectory, the arrow is the highest (vertical) height of the object during its fly over the slope.
		
		Args:
			gradient (float): the slope gradient (:math:`dx/dy`)
			
		Returns:
			float: :math:`x_{arrow}`
		"""
		return (gradient-self.B)/(2*self.A)



class Line(object):
	"""
	A line, described by the two constants :attr:`a` and :attr:`b` in :math:`z=ax + b`.
			
	Args:
		a (float): the rate of increase coefficient.
		b (float): the y-intercept.
	"""
	def __init__(self,S=None,a=None,b=None):
		"""
		Constructor: S or {a,b}
		
		Args:
			S (:class:`TwoD.Objects.Segment`): a segment from which to deduce the line.
		"""
		if(not S==None):
			self.a=S.slope_gradient
			self.b=S.points[0,1]-(self.a*S.points[0,0])
		else:
			self.a=a
			self.b=b

def get_roots(A,B,C):
	"""
	Gives the roots of the second order equation :math:`z=Ax^2 + Bx + C`. If there is no solutions, the returned array is empty. If the there is one solution only, it will be duplicated in the output array.
			
	Args:
		A (float): the square coefficient
		B (float): the linear coefficient
		C (float): the constant coefficient
	
	Returns:
		:class:`Common.Math.Vector2` [x_root1, x_root2]: array containing the two roots.
	"""
	D=B*B-4.*A*C
	if(D<0.):
		return np.array([])
	else:
		D=np.sqrt(D)
		return Math.Vector2([(-B+D)/(2.*A),(-B-D)/(2.*A)])

def get_line_parabola_intersections(L,P):
	"""
	Computes the intersections between a line and a parabola. If no solutions, return a NaN-filled array.
	
	Args:
		L (:class:`TwoD.Geoms.Line`): the line
		P (:class:`TwoD.Geoms.Parabola`): the parabola
	Returns:
		:class:`numpy.ndarray` [[x_i1,z_i1], [x_i2, z_i2]]: the two solutions.
	"""
	roots = get_roots(P.A,P.B-L.a,P.C-L.b)	#The roots are the x-coords of the intersection(s). They are the solutions of the 2nd order polynom coming from : parabola_equation = line_equation.
	if(len(roots)>0):	#If the solution(s) exists
		return np.array([ [roots[0],L.a*roots[0]+L.b], [roots[1],L.a*roots[1]+L.b] ])
	return np.array([[np.nan,np.nan],[np.nan,np.nan]])	#No solutions

def get_segment_parabola_intersections(S,P):
	"""
	Computes the intersection(s) between a segment and a parabola. If no solutions, return an empty array.
	
	Args:
		S (:class:`TwoD.Objects.Segment`): the segment
		P (:class:`TwoD.Geoms.Parabola`): the parabola
	Returns:
		:class:`numpy.ndarray` [[x_i1,z_i1], [x_i2, z_i2]]: the eventual solutions.
	"""
	potential_intersections=get_line_parabola_intersections(Line(S),P)
	valid_ids=[]
	for i in [0,1]:
		if(potential_intersections[i,0]>S.points[0,0] and potential_intersections[i,0]<S.points[1,0]):
			valid_ids.append(i)
	return potential_intersections[valid_ids]

#FIXME : remove
#def rotate(vector,angle):
	#"""
	#Gives the vector, rotated by the given angle in the trigonometric way
	
	#Args:
		#vector (:class:`numpy.ndarray` [x,z]): coordinates of the vector
		#angle (float): the angle in radians
	#Returns:
		#:class:`numpy.ndarray` [x,z]: the rotated vector
	#"""
	#c, s = np.cos(angle), np.sin(angle)
	#return np.array([[c, -s], [s, c]]).dot(vector)


def find_next_bounce(sim,rock,rock_parabola):
	"""
	In a given simulation, for a given rock, find the next rebound on the soil. The algorithm works as follows:
	
	#. compute the rock parabola
	#. loop on the segments from the :attr:`TwoD.Objects.Rock.current_segment` in the :attr:`TwoD.Objects.Rock.flying_direction`, and try to find a parabola-segment intersection
	#. keep the intersection with the soil if it is compatible with the rock flying direction and if it does not correspond to the previous rock rebound point
	
	Args:
		sim (:class:`TwoD.Simulations.Simulation`): the simulation
		rock (:class:`TwoD.Objects.Rock`): the rock
	Returns:
		:class:`Common.Math.Vector2` [x_i1,z_i1] ; :class:`TwoD.Objects.Segment` ; :class:`Parabola` : intersection point ; corresponding segment ; rock parabola
	"""
	Debug.Print("Find next bounce")
	trying_segt=rock.current_segment
	#LOOP ON SEGMENTS#
	while(True):
	#for l in range(0,10):
		Debug.Print("Trying to find rebound on segment n",trying_segt.index)
		intersections=get_segment_parabola_intersections(trying_segt,rock_parabola)
		Debug.Print("Potential intersections:",intersections)
		#LOOP ON POTENTIAL INTERSECTIONS#
		for i in intersections:
			delta_x=i[0]-rock.pos[0]
			if(np.sign(delta_x)==rock.flying_direction and abs(delta_x)>1e-7):
				Debug.Print("Keep intersection:",i,"with delta_x:",delta_x)
				return Math.Vector2(i),trying_segt #WE FOUND, SO STOP HERE
			else:
				Debug.Print("Invalid intersection:",i,"with delta_x:",delta_x)
				continue #GO TO THE NEXT INTERSECTION, IF ANY
		#NO VALID INTERSECTION, GO TO NEXT SEGMENT
		ID=trying_segt.index+rock.flying_direction
		if(ID<0 or ID>len(sim.terrain.segments)-1):
			raise("find_next_bounce(): no intersection could be found for this rock on the whole terrain.")
		trying_segt=sim.terrain.segments[ID]
		Debug.Print("No valid intersection found")



"""
Copyright (c) 2017, 2018, 2019 Irstea
Copyright (c) 2017, 2018, 2019 François Kneib
Copyright (c) 2017, 2018, 2019 Franck Bourrier
Copyright (c) 2017, 2018, 2019 David Toe
Copyright (c) 2017, 2018, 2019 Frédéric Berger
Copyright (c) 2017, 2018, 2019 Stéphane Lambert

This file is part of PlatRock.

PlatRock is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, version 3 of the License.

PlatRock is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Foobar. If not, see <http://www.gnu.org/licenses/>.
"""

"""
This module is used by the TwoD model. It is a kind of master module that handles simulations.
"""
#TODO:
# - find a workaround for rebounds that are too small


from . import Objects, Geoms
import platrock.Common.BounceModels as BounceModels
import platrock.Common.Debug as Debug
import platrock.Common.Math as Math
import platrock.Common.Outputs as Out
import platrock.Common.Simulations
import numpy as np
import time,os,zipfile
import sys,traceback


class Simulation(platrock.Common.Simulations.Simulation):
	"""
	A simulation
	
	Args:
		terrain (:class:`TwoD.Objects.Terrain`): the terrain of the simulation 
		gravity (float): the --- positive --- gravity acceleration value
		rocks (list [:class:`TwoD.Objects.Rock`]): list of rocks, automatically filled
		checkpoints (list [:class:`TwoD.Objects.Checkpoint`, ...]): all the checkpoints. Checkpoints data are post-computed with :meth:`update_checkpoints` from all contacts of all rocks.
		name (String): the name of the simulation, mandatory for the WebGUI
		rocks_start_params (dict): a dictionnary of the rocks start parameters (see Examples/2D.py for more details)
		status (String): available status: ["init"|"running"|"pause"|"finished"|"error"]
		
		nb_rocks (int): number of rocks to launch
		rocks_volume (float): volume of the rocks
		rocks_density (float): density of the rocks
		rocks_x (float): initial x corrdinate of the rocks
		rocks_z (float): initial height of the rocks
		rocks_vx (float): initial horizontal velocity of the rocks
		rocks_vz (float): initial vertical velocity of the rocks
		rocks_angVel (float): initial angular velocity of the rocks
		
		override_rebound_params (bool): if True, the rebound parameters will be overriden by :attr:`override_roughness`, :attr:`override_R_n`, :attr:`override_R_t`, :attr:`override_phi`, :attr:`override_v_half` and :attr:`override_bounce_model_number`
		override_forest_params (bool): if True, the forest parameters will be overriden by :attr:`override_enable_forest` :attr:`override_trees_density` and :attr:`override_trees_dhp_mean` and :attr:`override_trees_dhp_std`
		override_enable_forest (bool)
		override_bounce_model_number=None
		override_roughness=None
		override_R_n=None
		override_R_t=None
		override_phi=None
		override_v_half=None
		override_trees_density=None
		override_trees_dhp_mean=None
		override_trees_dhp_std=None
	"""
	webui_typename="PlatRock 2D"
	valid_rocks_start_params=np.array([
		["nb_rocks",		"Rocks count",						int,		1,		10000	],
		["rocks_volume",	"Rocks volume (m<sup>3</sup>)",		float,		0.1,	100.	],
		["rocks_vx",		"Horizontal velocity (m/s)",		float,		-10,	10,		],
		["rocks_vz",		"Vertical velocity (m/s)",			float,		-10,	10,		],
		["rocks_angVel",	"Angular velocity (rad/s)",			float,		-6.3,	6.3,	],
		["rocks_x",			"Horizontal position (m)",			float,		0 ,		np.inf,	],
		["rocks_z",			"Falling height (m)",				float,		0.1,	50,		],
		["rocks_density",	"Density (kg/m<sup>3</sup>)",		float,		500,	5000,	]
	])
	
	def __init__(self, checkpoints_x=[], rocks_start_params={}, **kwargs):
		super(Simulation, self).__init__(**kwargs)
		#FIXME : DOC
		"""
		Constructor
		
		Args:
			checkpoints_x (list [x0, x1, ...]): all the checkpoints x postions. Checkpoints data are post-computed with :meth:`update_checkpoints` from all contacts of all rocks.
		"""
		self.checkpoints=[Objects.Checkpoint(x) for x in checkpoints_x]
		
		# SET THE ROCKS DEFAULT PARAMS :
		self.nb_rocks=100
		self.rocks_volume=2.
		self.rocks_density=2650.
		self.rocks_x=1.
		self.rocks_z=5.
		self.rocks_vx=1.
		self.rocks_vz=0.
		self.rocks_angVel=0.
		#self.rocks_start_params=rocks_start_params
		#OVERRIDE THE DEFAULT PARAMS, IF AVAILABLE IN rocks_start_params dict.
		for k in rocks_start_params.keys():
			if(k in self.valid_rocks_start_params[:,0]):
				self.__dict__[k]=rocks_start_params[k]
		self.override_mu_r=30
		
	def save_to_file(self):
		"""
		Store (pickle to json) the simulation into a file whose path and name is the return result of :meth:`get_dir`. Note that the file will exclude all data that arises from calculations, this method is meant to save the simulation setup only.
		"""
		#store to local variables all data that shouldn't be saved to the file, then clear the corresponding simulation attributes :
		output=self.output
		self.output=None
		forest_impact_model=self.forest_impact_model
		self.forest_impact_model=None
		checkpoints=self.checkpoints[:]
		self.checkpoints=[Objects.Checkpoint(c.x) for c in checkpoints] #NOTE: keep the checkpoints position, drop the data
		if(self.terrain): #don't save the params zones as it's not necessary and jsonpickle don't deal with this weird types
			params_zones=self.terrain.params_zones
			self.terrain.params_zones=[]
		if hasattr(self,'hdf5_io'):
				hdf5_io = self.hdf5_io
				self.hdf5_io=None
		platrock.Common.Simulations.Simulation.save_to_file(self)	#actual file write is here. Don't use super as save_to_file is imported in TwoDShape.
		#restore the not-saved data:
		self.output=output
		self.forest_impact_model=forest_impact_model
		self.checkpoints=checkpoints
		if(self.terrain):
			self.terrain.params_zones=params_zones
		if hasattr(self,'hdf5_io'):
			self.hdf5_io=hdf5_io
		Debug.Print("... DONE")
	
	def update_checkpoints(self):
		"""
		Update all the simulation :class:`TwoD.Objects.Checkpoint` according to all contacts of all rocks. This is a post-treatement feature, it is supposed to be triggered after the simulation is terminated.
		"""
		#FIXME: there seems to be a bug when checkpoint position == contact pos, or initial rock pos.
		for chkP in self.checkpoints:
			chkP.init_data(self)
		for i in range(self.nb_rocks):
			contacts_pos=self.output.get_contacts_pos(i)
			for chkP in self.checkpoints:
				indices=np.where(contacts_pos[:,0]>=chkP.x)[0]
				if(len(indices)>0):
					chkP.rocks_ids.append(i)
					id_before=indices[0]-1
					pos=contacts_pos[id_before]
					vel=self.output.get_contacts_vels(i)[id_before]
					angVel=self.output.get_contacts_angVels(i)[id_before]
					if(self.output.get_contacts_types(i)[id_before] == Out.ROLL):
						chkP.heights.append(0.0)
						v0_square=Math.Vector2(vel).norm()**2
						vf_square=Math.Vector2(self.output.get_contacts_vels(i)[id_before+1]).norm()**2
						d_tot=Math.Vector2(contacts_pos[id_before+1]-pos).norm()
						dist_ratio = (chkP.x-contacts_pos[id_before][0])/(contacts_pos[id_before+1][0]-contacts_pos[id_before][0])
						d=d_tot*dist_ratio
						vel=np.sqrt( (vf_square-v0_square)/d_tot * d + v0_square )
						chkP.vels.append(vel*Math.Vector2(contacts_pos[id_before+1]-contacts_pos[id_before]).normalized())
						chkP.angVels.append(vel/((3*self.output.volumes[i]/4/np.pi)**(1/3)))
					elif self.output.get_contacts_types(i)[id_before] == Out.MOTION :
						dist_ratio = (chkP.x-contacts_pos[id_before][0])/(contacts_pos[id_before+1][0]-contacts_pos[id_before][0])
						chkP.heights.append(	contacts_pos[id_before][1]*(1-dist_ratio) + contacts_pos[id_before+1][1]*dist_ratio - chkP.base_height)
						chkP.vels.append(		vel*(1-dist_ratio) + self.output.get_contacts_vels(i)[id_before+1]*dist_ratio)
						chkP.angVels.append(	angVel*(1-dist_ratio) + self.output.get_contacts_angVels(i)[id_before+1]*dist_ratio)
					else:
						parabola=Geoms.Parabola(pos=Math.Vector2(pos),vel=Math.Vector2(vel),g=self.gravity)
						chkP.heights.append(parabola.A*chkP.x**2 + parabola.B*chkP.x + parabola.C - chkP.base_height)
						chkP.vels.append(vel.copy()) #make a copy as we modify the value below, and don't want to modify the s.output.contacts arrays
						chkP.vels[-1][1]=-self.gravity*(chkP.x-pos[0])/vel[0] + vel[1]
						chkP.angVels.append(angVel)
		for chkP in self.checkpoints:
			chkP.crossings_ratio=len(chkP.heights)/self.nb_rocks
		self.output.checkpoints=self.checkpoints
	
	def get_stops_cdf(self,nbins=200):
		xmax=np.zeros(self.output._rocks_counter)
		if (self.output is not None) and self.output._rocks_counter>0 :
			for r_id in range(0,self.output._rocks_counter):
				xmax[r_id] = self.output.get_contacts_pos(r_id)[:,0].max()
		xmax.sort()
		rocks_out_of_bounds_count=np.sum(xmax>self.terrain.get_points()[-1,0]) 
		hist,bin_edges=np.histogram(xmax,np.linspace(self.terrain.segments[0].points[0][0],self.terrain.segments[-1].points[1][0],nbins))
		hist=np.cumsum(hist)
		hist=np.append(hist,hist[-1])
		hist=hist/(hist.max()+rocks_out_of_bounds_count)
		out=np.zeros([2,len(hist)])
		out[0,:]=bin_edges
		out[1,:]=hist
		return out
		
	
	def results_to_zipfile(self,filename=None):
		"""
		Create a zip file into the folder returned by :meth:`get_dir` named results.zip containing two text files. "stops.csv" contains info about rocks end position, and "checkpoints.csv" contains the checkpoints infos.
		"""
		if(filename is None):
			filename=self.get_dir()+"results.zip"
		zf = zipfile.ZipFile(filename, "w")
		output_str="volume;x_stop\n"
		for i in range(self.nb_rocks):
			output_str+=str(self.output.volumes[i])+";"+str(self.output.get_contacts_pos(i)[-1,0])+"\n"
		zf.writestr("stops.csv",output_str)
		
		output_str="x_checkpoint;vx;vz;volume;height;angVel;Ec_t;Ec_r\n"
		for chckpt in self.checkpoints:
			for i in range(len(chckpt.heights)):
				rock_id=chckpt.rocks_ids[i]
				mass=self.output.volumes[rock_id]*self.output.densities[rock_id]
				output_str+=str(chckpt.x)+";"
				output_str+=str(chckpt.vels[i][0])+";"
				output_str+=str(chckpt.vels[i][1])+";"
				output_str+=str(self.output.volumes[rock_id])+";"
				output_str+=str(chckpt.heights[i])+";"
				output_str+=str(chckpt.angVels[i])+";"
				output_str+=str(0.5*mass*Math.Vector2(chckpt.vels[i]).norm()**2)+";"
				output_str+=str(0.5*self.output.inertias[rock_id]*chckpt.angVels[i]**2)+"\n"
		zf.writestr("checkpoints.csv",output_str)
		
		if(os.path.isfile(self.get_dir()+"terrain_overview.pdf")):
			zf.write(self.get_dir()+"terrain_overview.pdf",arcname="trajectories_overview.pdf")
		zf.close()
	
	def before_run_tasks(self):
		super(Simulation, self).before_run_tasks()
		if(self.enable_forest and (not self.terrain.forest_available)):
			self.override_forest_params=True
		if(self.override_forest_params and self.override_trees_density==0):
			self.enable_forest=False
		if(self.enable_forest and self.forest_impact_model==None):
			self.forest_impact_model=BounceModels.Toe_Tree()
	
	def before_rock_launch_tasks(self):
		super(Simulation, self).before_rock_launch_tasks()
		self.current_rock=Objects.Rock(x=self.rocks_x, height=self.rocks_z, vel=Math.Vector2([self.rocks_vx,self.rocks_vz]), angVel=self.rocks_angVel, volume=self.rocks_volume, density=self.rocks_density)
		self.current_rock.update_current_segment(self.terrain)
		self.current_rock.pos[1]+=self.current_rock.current_segment.get_y(self.current_rock.pos[0])
		self.output.add_rock(self.current_rock)
		self.output.add_contact(self.current_rock,Math.Vector2([0.,1.]),Out.START) #store the rock initial position
		Debug.Print("New rock:",self.current_rock.pos,self.current_rock.vel)
	
	def after_rock_propagation_tasks(self):
		type(self).__bases__[0].after_rock_propagation_tasks(self)
		if(self.current_rock.out_of_bounds):
			if (not platrock.SICONOS_FOUND) or (type(self)!=platrock.TwoDShape.Simulations.Simulation):
				#Fly into the void...
				parab=Geoms.Parabola(self.current_rock,g=self.gravity)
				terrain_Dx=self.terrain.get_points()[-1][0] - self.terrain.get_points()[0][0]
				if(self.current_rock.vel[0]>0):
					arrival_x=self.terrain.get_points()[-1][0]+0.05*terrain_Dx
				else:
					arrival_x=self.terrain.get_points()[0][0]-0.05*terrain_Dx
				self.current_rock.fly(Math.Vector2([arrival_x,parab.get_value_at(arrival_x)]),self,self.current_rock.current_segment)
			self.output.add_contact(self.current_rock,Math.Vector2([0.,0.]),Out.OUT)
		else:
			#Mark the stop position
			self.output.add_contact(self.current_rock,Math.Vector2([0.,0.]),Out.STOP)
	
	def after_successful_run_tasks(self):
		type(self).__bases__[0].after_successful_run_tasks(self)
		self.update_checkpoints()
		if platrock.web_ui:
			import platrock.GUI.Plot2D as Plot2D
			with open(self.get_dir()+'terrain_overview_plotly.html','w') as f:
				if platrock.SICONOS_FOUND and isinstance(self,platrock.TwoDShape.Simulations.Simulation):
					f.write(Plot2D.get_plotly_raw_html(self,50))
				else:
					f.write(Plot2D.get_plotly_raw_html(self,100))
			self.results_to_zipfile()

	def rock_propagation_tasks(self):
		r=self.current_rock
		Debug.Print("\nNew rock propagation loop")
		if(abs(r.vel[0])<1e-5):
			r.vel[0]=np.sign(r.vel[0])*1e-5 #the horizontal velocity can't be zero as it would not create a parabolic trajectory
			if(r.vel[0]==0): #if the vel is exactly 0
				r.vel[0]=1e-5
			if len(self.output.get_contacts_types(-1)) == 1 and self.output.get_contacts_types(-1)[0] == Out.START :
				self.output.get_contacts_vels(-1)[0,0]=1e-5
			Debug.warning("The rock vel along x is too small (<1e-5), set it to",r.vel[0])
		
		#FOREST=======================================================================================================#
		if(self.enable_forest):
			loop_counter=-1
			rolling = False
			r.force_roll=False
			# Loop on portions of the trajectory. The loop ends only when the rock rebounds on the SOIL 
			while(True):
				#INIT:
				loop_counter+=1
				Debug.Print("Start tree loop")
				r.update_flying_direction() #if the previous loop changed the rock propagation direction
				parabola=Geoms.Parabola(r,g=self.gravity)
				Debug.Print("Rock parabola:",parabola.A,"*x*x +",parabola.B,"*x +",parabola.C)
				arrow_x=parabola.get_arrow_x_from_gradient(r.current_segment.slope_gradient) #what will be the x coord of the maximal height reached by the rock on this segment ?
				arrow_height=parabola.get_value_at(arrow_x)-r.current_segment.get_y(arrow_x)
	
				#DO WE ROLL OR DO WE FLY ?
				if loop_counter==0: #this means that the rock is currently in contact with the SOIL (no tree, OR not in flight, OR first portion of roll in this segment). Roll can only be triggered in this case.
					if arrow_height < r.radius/10 or r.force_roll:
						Debug.info("Set rolling to True with arrow_x=",arrow_x,"and arrow_height=",arrow_height)
						rolling=True
						self.output.del_contact(-1,-1) #remove the last contact, we will add a rolling contact instead.

				#HORIZONTAL DISTANCE TRAVELLED ON THE CURRENT SEGMENT WITH THE ROLL
				if rolling :
					mu_r = self.override_mu_r if self.override_rebound_params else r.current_segment.mu_r
					AR=BounceModels.Azzoni_Roll(r.vel,mu_r,-r.current_segment.slope,self.gravity,start_pos=r.pos,tan_slope=-r.current_segment.slope_gradient,A=r.A)
					self.output.add_contact(r,r.current_segment.normal,Out.ROLL)
					arrival_point=AR.stop_pos
					#Limit the arrival point to the segment extremas
					if arrival_point[0] < r.current_segment.points[0][0]:
						arrival_point=r.current_segment.points[0]
					elif arrival_point[0] > r.current_segment.points[1][0]:
						arrival_point=r.current_segment.points[1]
					else:
						AR.until_stop=True
					arrival_point=Math.Vector2(arrival_point)
					xdist_travel_on_segment=arrival_point[0]-r.pos[0]
				
				#HORIZONTAL DISTANCE TRAVELLED ON THE CURRENT SEGMENT WITH THE PARABOLA
				else:
									#PARABOLA COMPUTATION, STOP CONDITION:
					try:
						freefall_bounce_point,rebound_segment=Geoms.find_next_bounce(self,r,rock_parabola=parabola) #free fall bounce point, previsional bounce point if we don't consider the forest
					except :
						r.out_of_bounds=True
						r.is_stopped=True #just to stop propagation in main loop (Common.Simulations.run())
						Debug.info("No intersection could be found, check rock.out_of_bounds to find out what happened.")
						return
					if(r.flying_direction>0):	# The end of the trajectory on the current segment is determined either by the next rebound or by the segment ends.
						xdist_travel_on_segment=min(freefall_bounce_point[0],r.current_segment.points[1][0])-r.pos[0]
					else:
						xdist_travel_on_segment=max(freefall_bounce_point[0],r.current_segment.points[0][0])-r.pos[0]

				#FOREST PARAMETERS OVERRIDES:
				if(self.override_forest_params):
					trees_density=self.override_trees_density
					dhp_std=self.override_trees_dhp_std
					dhp_mean=self.override_trees_dhp_mean
				else:
					trees_density=r.current_segment.trees_density
					dhp_std=r.current_segment.dhp_std
					dhp_mean=r.current_segment.dhp_mean
				if(trees_density>1e-5 and dhp_mean>1e-5 and dhp_std>1e-5):
					dhp=Math.get_random_value_from_gamma_distribution(dhp_mean,dhp_std,self.random_generator)
				else:
					trees_density=0
					dhp=0
				
				#WHAT DISTANCE CAN WE STATISTICALLY TRAVEL ALONG X BEFORE REACHING A TREE ?
				if(trees_density<=1e-5 or dhp<0.01): # avoid zero division
					next_tree_impact_xdist=np.inf
				else:
					next_tree_impact_xdist=((100./np.sqrt(trees_density))**2)/(dhp_mean/100. + r.radius*2.)
				Debug.Print("Mean next_tree_impact_xdist :",next_tree_impact_xdist)
				#ADD LINEAR RANDOMNESS BETWEEN 0 (=REACH A TREE NOW) AND 2*next_tree_impact_xdist (TRAVEL 2 TIMES THE STATISTICAL MEAN DISTANCE)
				next_tree_impact_xdist*=2.*self.random_generator.rand()
				
				#WILL THERE BE A TREE-CONTACT BEFORE REACHING THE END OF THIS SEGMENT ? ...
				Debug.Print("Comparing abs(xdist_travel_on_segment)=abs(",xdist_travel_on_segment, ") with next_tree_impact_xdist=",next_tree_impact_xdist)
				if(next_tree_impact_xdist<abs(xdist_travel_on_segment)): # then an impact with a tree will hapen before the end of the travel on this segment.
				#... YES! ROLL OR FLY TO THIS POINT IN ANY CASES
					impact_point=Math.Vector2([r.pos[0]+r.flying_direction*next_tree_impact_xdist,0.0]) # set x first
					if rolling:
						AR.until_stop=False #finally the rock will not roll until stop, as there is a tree before.
						impact_point[1]=r.current_segment.get_y(impact_point[0]) # set z
						r.roll(self,AR,impact_point)
					else:
						impact_point[1]=parabola.get_value_at(impact_point[0]) # set z
						impact_height = impact_point[1]-r.current_segment.get_y(impact_point[0])
						r.fly(impact_point,self,r.current_segment)
						#IF THE ROCK FLIES LOW ENOUGH, MODIFY ITS VEL AND STORE THE CONTACT:
						if(impact_height>2): #no rock-tree impact if impact height is too high
							continue #the while(True) loop
					#After roll or fly, do the rock-tree contact:
					vx,vy,vz=self.forest_impact_model.get_output_vel(r.vel.norm(),r.volume,self.random_generator.rand(),dhp,np.sign(r.vel[1])*abs(np.degrees(np.arctan(r.vel[1]/r.vel[0]))))
					vx=vx*np.sign(r.vel[0])
					toe_output_vel=Math.Vector2([vx,vz])
					#Linear interpolation of velocity norm between PlatRock input vel and Toe's input one :
					out_vel_norm=r.vel.norm()*toe_output_vel.norm()/self.forest_impact_model.last_computed_data["input_vel"]
					toe_output_vel=toe_output_vel.normalized()*out_vel_norm
					if rolling: #if the rock was rolling, also roll after contact
						r.vel=Math.normalized2(r.current_segment.branch)*out_vel_norm*np.sign(vx)
						self.output.add_contact(r,(1*r.flying_direction,0.),Out.ROLL_TREE)
					else: #if the rock was flying, also do a fly after contact
						r.vel[0]=toe_output_vel[0]
						r.vel[1]=toe_output_vel[1]
						self.output.add_contact(r,(1*r.flying_direction,0.),Out.TREE)
					Debug.Print("Impact with a tree at:",impact_point,", dhp=",dhp,"cm", "output vel is",r.vel)
				#... NO! THERE WILL BE A FREEFLIGHT OR ROLL
				else:
					#ROLL TO THE NEXT SEGMENT OR UNTIL STOP:
					if rolling:
						# as r.current_segment will be modified in r.roll():
						prev_segment_id=r.current_segment.index
						r.roll(self,AR,arrival_point)# note: AR.until_stop have been set earlier
						#if we rolled until out_of_bounds occured. Note that r.is_stopped==True here.
						if r.out_of_bounds :
							self.output.add_contact(r,Math.Vector2([0.,0.]),Out.SOIL)
						#handle junction with next segment
						if not r.is_stopped :
							angle_segs=self.terrain.get_angle_between(prev_segment_id, r.current_segment.index) #the slope discontinuity angle
							if abs(angle_segs-np.pi) < 1e-2 : # angle=180° (collinear)
								Debug.info("After roll, collinear segment -> still roll.")
								r.force_roll=True
								self.output.add_contact(r,Math.Vector2([0.,1.]),Out.ROLL) # the next loop will be a roll, so this contact will be immediately removed
							elif angle_segs>np.pi:	# angle>180°, just a soil-like contact so that at next loop we will have a flight
								Debug.info("After roll, free fly.")
								self.output.add_contact(r,Math.Vector2([0.,1.]),Out.SOIL)
							elif angle_segs>np.pi/2:	# 90°>angle>180°, we have a real contact with the segment, compute the bounce here.
								Debug.info("After roll, contact on soil.")
								self.output.add_contact(r,r.bounce(self,r.current_segment,disable_roughness=True),Out.SOIL)
								r.update_flying_direction()
								r.update_current_segment(self.terrain) #the previous r.bounce may have changed the rock vel x-direction. Update the current_segment, knowing that the rock is exactly between two segments.
							else:	# angle<90°, we have a frontal impact, stop the rock.
								Debug.info("After roll, stop rock.")
								r.is_stopped=True
						break #end the infinite loop

					#FLY TO THE NEXT SEGMENT OR THE NEXT SOIL CONTACT
					else:
						#i) the rock will fly until a rebound on the current segment :
						if(rebound_segment.index==r.current_segment.index):
							Debug.Print("No tree impact detected until next bounce on soil")
							r.fly(freefall_bounce_point,self,rebound_segment) #fly to the bounce point
							self.output.add_contact(r,r.bounce(self, rebound_segment),Out.SOIL) #bounce here
							break	#end the infinite loop
						#ii) the rock will fly to the next segment
						else:
							#do a PORTION of flight, until reaching the next segment
							Debug.Print("Fly above the current segment to the begining of the next one")
							if(r.flying_direction>0):
								next_x=r.current_segment.points[1][0]
								next_seg=self.terrain.segments[r.current_segment.index+1]
							else:
								next_x=r.current_segment.points[0][0]
								next_seg=self.terrain.segments[r.current_segment.index-1]
							r.fly(Math.Vector2([next_x,parabola.get_value_at(next_x)]),self,next_seg)
		#/FOREST=======================================================================================================#
		
		#NO FOREST=====================================================================================================#
		else: #no forest :
			r.update_flying_direction() #if the previous loop changed the rock propagation direction
			parabola=Geoms.Parabola(r,g=self.gravity)
			Debug.Print("Rock parabola:",parabola.A,"*x*x +",parabola.B,"*x +",parabola.C)
			arrow_x=parabola.get_arrow_x_from_gradient(r.current_segment.slope_gradient) #what will be the x coord of the maximal height reached by the rock on this segment ?
			arrow_height=parabola.get_value_at(arrow_x)-r.current_segment.get_y(arrow_x)
			
			# FLIGHT #
			if arrow_height > r.radius/10 and not r.force_roll : #the arrow is high-enough to do a FLIGHT:
				Debug.info("FLY")
				try:
					bounce_point,rebound_segment=Geoms.find_next_bounce(self,r,parabola) #free fall bounce point, previsional bounce point
				except : #no bounce point have been found...
					r.out_of_bounds=True
					r.is_stopped=True
					Debug.warning("No intersection could be found, maybe the rock went out of terrain bounds ?")
					return
				r.fly(bounce_point,self,rebound_segment)
				if not r.is_stopped :
					#Bounce on the destination segment :
					self.output.add_contact(r,r.bounce(self, rebound_segment),Out.SOIL)

			# ROLL #
			else:
				Debug.info("ROLL")
				mu_r = self.override_mu_r if self.override_rebound_params else r.current_segment.mu_r
				AR=BounceModels.Azzoni_Roll(r.vel,mu_r,-r.current_segment.slope,self.gravity,start_pos=r.pos,tan_slope=-r.current_segment.slope_gradient,A=r.A)
				#As we are rolling, replace the last contact recorded by a roll-contact
				self.output.del_contact(-1,-1) #remove the last contact, we will add a rolling contact instead.
				self.output.add_contact(r,r.current_segment.normal,Out.ROLL)
				arrival_point=AR.stop_pos
				#Limit the arrival point to the segment extremas
				if arrival_point[0] < r.current_segment.points[0][0]:
					arrival_point=r.current_segment.points[0]
				elif arrival_point[0] > r.current_segment.points[1][0]:
					arrival_point=r.current_segment.points[1]
				else:
					AR.until_stop=True
				arrival_point=Math.Vector2(arrival_point)
				# as r.current_segment will be modified in r.roll()
				prev_segment_id=r.current_segment.index 
				r.roll(self,AR,arrival_point)
				#if we rolled until out_of_bounds occured. Note that r.is_stopped==True here.
				if r.out_of_bounds :
					self.output.add_contact(r,Math.Vector2([0.,0.]),Out.SOIL)
				#handle junction with next segment
				r.force_roll=False
				if not r.is_stopped :
					angle_segs=self.terrain.get_angle_between(prev_segment_id, r.current_segment.index) #the slope discontinuity angle
					if abs(angle_segs-np.pi) < 1e-2 : #collinear
						Debug.info("After roll, collinear segment -> still roll.")
						r.force_roll=True
						self.output.add_contact(r,Math.Vector2([0.,1.]),Out.ROLL) # the next loop will be a roll, so this contact will be immediately removed
					elif angle_segs>=np.pi :	# angle>180°, just a soil-like contact so that at next loop we will have a flight
						Debug.info("After roll, free fly.")
						self.output.add_contact(r,Math.Vector2([0.,1.]),Out.SOIL)
					elif angle_segs>np.pi/2:	# 90°>angle>180°, we have a real contact with the segment, compute the bounce here.
						Debug.info("After roll, contact on soil.")
						self.output.add_contact(r,r.bounce(self,r.current_segment,disable_roughness=True),Out.SOIL)
						r.update_flying_direction()
						r.update_current_segment(self.terrain) #the previous r.bounce may have changed the rock vel x-direction. Update the current_segment, knowing that the rock is exactly between two segments.
					else:	# angle<90°, we have a frontal impact, stop the rock.
						Debug.info("After roll, stop rock.")
						r.is_stopped=True



"""
Copyright (c) 2017, 2018, 2019 Irstea
Copyright (c) 2017, 2018, 2019 François Kneib
Copyright (c) 2017, 2018, 2019 Franck Bourrier
Copyright (c) 2017, 2018, 2019 David Toe
Copyright (c) 2017, 2018, 2019 Frédéric Berger
Copyright (c) 2017, 2018, 2019 Stéphane Lambert

This file is part of PlatRock.

PlatRock is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, version 3 of the License.

PlatRock is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Foobar. If not, see <http://www.gnu.org/licenses/>.
"""


"""
This module gather all simulations common things for ThreeD and TwoD models.
"""
import time,traceback,yappi
import numpy as np
import subprocess,os,sys
from . import BounceModels
import platrock.Common.Outputs
import platrock.Common.Debug as Debug
import jsonpickle
import jsonpickle.ext.numpy as jsonpickle_numpy
jsonpickle_numpy.register_handlers()
jsonpickle.set_encoder_options('simplejson', sort_keys=True, indent=1)

class Simulation(object):
	"""
	"""
	def __init__(self, terrain=None, gravity=9.8, name=None, project=None, enable_forest=False, random_seed=None):
		#FIXME : DOC
		"""
		"""
		t=time.localtime()
		if(name==None):self.name="New simulation ("+str(t.tm_mday).rjust(2,'0')+"-"+str(t.tm_mon).rjust(2,'0')+"-"+str(t.tm_year)+", "+str(t.tm_hour).rjust(2,'0')+":"+str(t.tm_min).rjust(2,'0')+":"+str(t.tm_sec).rjust(2,'0')+")"
		else:self.name=str(name)
		self.terrain=terrain
		self.gravity=gravity
		self.project=project
		self.enable_forest=enable_forest
		self.status="init"
		self.forest_impact_model=None
		self.nb_rocks=0
		self.current_rock=None
		self.benchmark=False
		self.checkpoints=None
		self.output=None
		self.current_rock_number=-1
		
		#OVERRIDE FOREST PARAMETERS :
		self.override_forest_params=False
		self.override_trees_density=0
		self.override_trees_dhp_mean=30
		self.override_trees_dhp_std=10
		#OVERRIDE SEGMENTS PARAMETERS :
		self.override_rebound_params=False
		self.override_bounce_model_number=0
		self.override_roughness=0.1
		self.override_R_n=0.8
		self.override_R_t=0.7
		self.override_phi=30
		self.override_v_half=5
		
		self.number_to_model_correspondance = {}
		
		if(random_seed==None):
			self.random_generator=np.random.RandomState()
		else:
			self.random_generator=np.random.RandomState(random_seed)
	
	def get_full_path(self):
		"""
		Get the full path of the simulation pickle file, composed of: "/the_project_path/simulation_name/setup.pickle"
		
		Returns:
			String: "/the_project_path/simulation_name/setup.pickle"
		"""
		if(self.project):
			return self.project.path+"/"+self.name+"/setup.json"
		else:
			Debug.info("This simulation has no project.")
			return "./"+self.name+"/setup.json"
	
	def get_dir(self):
		"""
		Get the path corresponding to the simulation
		
		Returns:
			String: "/the_project_path/simulation_name/"
		"""
		return self.get_full_path().rsplit("/",1)[0]+"/"
	
	def save_to_file(self):
		"""
		FIXME: DOC. Hint: this method will write any simulation into a json file with jsonpickle. However it is not intented to be used as it, rather call it from the simulation subclass after having removed the unwanted heavy attributes.
		"""
		Debug.Print("Save simulation...")
		prefix=self.get_dir()
		if(not os.path.isdir(prefix)):
			subprocess.call(["mkdir",prefix])
		with open(self.get_full_path(), "w", encoding="utf8") as f:
			project=self.project
			self.project=None
			f.write(jsonpickle.encode(self))
		self.project=project
	
	def before_run_tasks(self):
		Debug.add_logger(self.get_dir()+"log.txt")
		self.current_rock_number=-1
		#initialize the bounce models:
		for number in BounceModels.number_to_model_correspondance.keys():
			self.number_to_model_correspondance[number]=BounceModels.number_to_model_correspondance[number](self) # instanciation of each model
		self.output=platrock.Common.Outputs.Output(self)
		if(self.benchmark):
			yappi.start(builtins=True)
		self.status="running"
	
	def before_rock_launch_tasks(self):
		Debug.info("\n-----------------------------------\n.....::::: NEW ROCK N"+str(self.current_rock_number)+":::::.....\n-----------------------------------\n")
	
	def rock_propagation_tasks(self):
		return
	
	def after_rock_propagation_tasks(self):
		return
	
	def after_successful_run_tasks(self):
		self.status="finished"
	
	def after_failed_run_tasks(self):
		self.status="error"
		Debug.error(traceback.format_exc())
		
	def after_all_tasks(self):
		if(self.benchmark):
			yappi.stop()
			fs=yappi.get_func_stats()
			fs.sort('ttot')
			fs.print_all( columns={0: ('name', 36), 1: ('ncall', 8), 2: ('ttot', 8), 3: ('tsub', 8), 4: ('tavg', 8)} )
		if platrock.web_ui:
			self.save_to_file()
			Debug.Print("Simulation located in "+self.get_full_path()+" finished with status "+self.status+".")
		else:
			Debug.Print("Simulation finished with status "+self.status+".")
	
	def run(self,GUI=False):
		try:
			self.before_run_tasks()
		except:
			self.after_failed_run_tasks()
			self.after_all_tasks()
			return
		if(GUI):	#ThreeD GUI only
			self.GUI_enabled=True
			import platrock.GUI.View3D as View3D
			self.GUI=View3D
			if(self.terrain):self.GUI.draw_rock(self.terrain)
			if(self.enable_forest):self.GUI.draw_trees(self)
			self.GUI.V.ren.ResetCamera()
		try:
			while self.current_rock_number<self.nb_rocks-1 :
				self.current_rock_number+=1
				self.before_rock_launch_tasks()
				while( not self.current_rock.is_stopped):
					self.rock_propagation_tasks()
				self.after_rock_propagation_tasks()
			self.after_successful_run_tasks()
		except KeyboardInterrupt:
			pass
		except Exception:
			self.after_failed_run_tasks()
		finally:
			self.after_all_tasks()



"""
Copyright (c) 2017, 2018, 2019 Irstea
Copyright (c) 2017, 2018, 2019 François Kneib
Copyright (c) 2017, 2018, 2019 Franck Bourrier
Copyright (c) 2017, 2018, 2019 David Toe
Copyright (c) 2017, 2018, 2019 Frédéric Berger
Copyright (c) 2017, 2018, 2019 Stéphane Lambert

This file is part of PlatRock.

PlatRock is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, version 3 of the License.

PlatRock is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Foobar. If not, see <http://www.gnu.org/licenses/>.
"""


"""
This module is used by the ThreeD model. It handles all the bounce model types
"""
import numpy as np
from . import Debug,Math
import os

#FIXME: DOC in the whole file
class BounceModel(object):
	"""
	
	.. _bounce_params:
	
	Parent class of every bounce models. The table below represents the bounce-models dependency to the parameters. The soil parameters of the terrain elements (segments or faces) MUST be set regarding to this table. Similarly, the input files (csv or geojson) must be consistent with this table. Note that the parameters are case-sensitive.
	
	+------------------+---------------------------+-----------------+------------+------------+--------------+------------+
	|                  |:attr:`bounce_model_number`|:attr:`roughness`|:attr:`R_n` | :attr:`R_t`|:attr:`v_half`|:attr:`phi` |
	+------------------+---------------------------+-----------------+------------+------------+--------------+------------+
	|Classical         |            *              |        *        |     *      |     *      |              |            |
	+------------------+---------------------------+-----------------+------------+------------+--------------+------------+
	|Pfeiffer          |            *              |        *        |     *      |     *      |              |            |
	+------------------+---------------------------+-----------------+------------+------------+--------------+------------+
	|Bourrier          |            *              |        *        |            |     *      |      *       |     *      |
	+------------------+---------------------------+-----------------+------------+------------+--------------+------------+
	
	"""
	def __init__(self,simulation):
		self.updated_normal=None
		self.simulation=simulation
		if("TwoD" in self.simulation.__module__):
			self.set_updated_normal=self.set_updated_normal2D
			self.check_output_vel=self.check_output_vel_2D
		else:
			self.set_updated_normal=self.set_updated_normal3D
			self.check_output_vel=self.check_output_vel_3D
		
	def set_updated_normal2D(self,rock,normal,roughness):
		roughness_angle=self.simulation.random_generator.rand()*np.arctan(roughness/rock.radius)
		roughness_angle*=np.sign(rock.vel.cross(normal))
		self.updated_normal=normal.rotated(roughness_angle)
		if(self.updated_normal[1]<0.01):	#AVOID THE ROUGHNESS TO CAUSE SLOPE HIGHER THAN 90° OR LESS THAN -90°.
			Debug.info("Roughness caused slope higher than 90 or lower than -90 with normal =",self.updated_normal)
			self.updated_normal[1]=0.01		#Set the z component of the normal=0.01
			self.updated_normal[0]=np.sqrt(1-0.01**2)*np.sign(self.updated_normal[0])	#ajust the x component according to the normalization condition
		Debug.info("updated_normal =",self.updated_normal)
		
	def set_updated_normal3D(self,rock,normal,roughness):
		nd=(rock.vel-rock.vel.dot(normal)*normal).normalized()
		nt=(nd.cross(normal)).normalized()
		roughness_angle=self.simulation.random_generator.rand()*np.arctan(roughness/rock.radius)
		braking_ratio=self.simulation.random_generator.rand()
		deviation_ratio=np.sign(self.simulation.random_generator.rand()-0.5)*(1-braking_ratio)
		self.updated_normal=normal.rotated(roughness_angle*braking_ratio,nt)
		self.updated_normal=self.updated_normal.rotated(roughness_angle*deviation_ratio,nd)
		#FIXME: can the output vel be "inside" the terrain with that ?
	
	def get_velocity_decomposed(self,v):
		vn = v.dot(self.updated_normal)*self.updated_normal
		return [vn,v-vn]
	
	def check_output_vel_2D(self,r,f):
		if(r.vel.cross(f.branch)>-1e-10):
			direction=np.sign(r.vel[0])
			v_norm=r.vel.norm()
			angle = np.pi/180 if direction>0 else np.pi*(1-1/180)
			r.vel=f.branch.rotated(angle).normalized()*v_norm
			self.updated_normal=Math.Vector2([0,0]) #updated_normal is no longer meaningful and difficult to compute for complex rebound models -> disable it.
			Debug.info("Output vel is not valid (inside terrain), modify it to",r.vel)
	
	def check_output_vel_3D(self,r,f):
		pass

class Classical(BounceModel):
	"""
	Classical bounce model type.
	"""
	def __init__(self,*args):
		super(Classical, self).__init__(*args)
	def run(self,r,f,disable_roughness=False):
		"""
		FIXME:  DOC
		"""
		if(self.simulation.override_rebound_params):
			roughness=self.simulation.override_roughness
			R_n=self.simulation.override_R_n
			R_t=self.simulation.override_R_t
		else:
			roughness=f.roughness
			R_n=f.R_n
			R_t=f.R_t
		if(disable_roughness):
			roughness=0
		
		#APPLY REBOUND :
		Debug.Print("CLASSICAL REBOUND on",f)
		self.set_updated_normal(r,f.normal,roughness)
		Vn, Vt = self.get_velocity_decomposed(r.vel)
		Debug.Print("Old vel:",r.vel, "Vn,Vt=",Vn,Vt)
		r.vel=-Vn*R_n+Vt*R_t
		Debug.Print("New vel:",r.vel)
		self.check_output_vel(r,f)

class Pfeiffer(BounceModel):
	"""
	Pfeiffer bounce model type.
	"""
	def __init__(self,*args):
		super(Pfeiffer, self).__init__(*args)
	def run(self,r,f,disable_roughness=False):
		"""
		Runs the bounce model:
		
		#. randomly compute the LSA
		#. compute the rock velocity in the local slope coordinate system
		#. compute the friction function and the scaling factor
		#. modify the velocity
		#. finally rotate back the velocity into the global coordinates and apply it to the rock.
		
		Args:
			r (:class:`TwoD.Objects.Rock`): the rock that bounces
			s (:class:`TwoD.Objects.Segment`): the segment on which the rock bounces
		"""
		
		#SET PARAMS :
		if(self.simulation.override_rebound_params):
			roughness=self.simulation.override_roughness
			R_n=self.simulation.override_R_n
			R_t=self.simulation.override_R_t
		else:
			roughness=f.roughness
			R_n=f.R_n
			R_t=f.R_t
		if(disable_roughness):
			roughness=0
		
		#APPLY REBOUND :
		Debug.Print("PFEIFFER REBOUND on",f)
		self.set_updated_normal(r,f.normal,roughness)
		Vn, Vt = self.get_velocity_decomposed(r.vel)
		
		Ff=R_t+(1.-R_t)/ ((((Vt.norm()-r.angVel.norm()*r.radius)/6.096)**2) +1.2)	#Friction function 6.096=const (vitesse)
		SF=R_t/ ((Vn.norm()/(76.2*R_n))**2+1.)	#Scaling factor
		
		Vn=-Vn.normalized()*Vn.norm()*R_n/(1.+(Vn.norm()/9.144)**2)
		Vt=Vt.normalized()*np.sqrt( ( r.radius**2*(r.I*r.angVel.norm()**2+r.mass*Vt.norm()**2)*Ff*SF ) / ( r.I+r.mass*r.radius**2 ) )
		
		r.vel=Vn+Vt
		r.angVel=self.updated_normal.cross(Vt)   /   r.radius
		Debug.Print("New vel:",r.vel)
		Debug.Print("New angVel:",r.angVel)
		self.check_output_vel(r,f)

class Bourrier(BounceModel):
	"""
	Bourrier bounce model type.
	"""
	def __init__(self,*args):
		super(Bourrier, self).__init__(*args)
	def run(self,r,f,disable_roughness=False):
		"""
		
		"""
		
		#SET PARAMS :
		if(self.simulation.override_rebound_params):
			roughness=self.simulation.override_roughness
			R_t=self.simulation.override_R_t
			v_half=self.simulation.override_v_half
			phi=self.simulation.override_phi
		else:
			roughness=f.roughness
			R_t=f.R_t
			v_half=f.v_half
			phi=f.phi
		if(disable_roughness):
			roughness=0
		
		#APPLY REBOUND :
		Debug.Print("BOURRIER REBOUND on",f)
		self.set_updated_normal(r,f.normal,roughness)
		Vn, Vt = self.get_velocity_decomposed(r.vel)
		
		#FIXME: compute Vn.norm() and Vt.norm() once for perfomance improvement.
		R_n = v_half/(abs(Vn.norm())+v_half)
		psi = np.degrees(np.arctan((2.*(R_t*Vt.norm()+r.radius*R_t*r.angVel.norm()))/(7.*Vn.norm()*(1.+R_n))))
		if (abs(psi)<phi):
			Debug.Print("Adhesion bounce")
			a,b=(5./7.)*R_t*Vt.norm(), (2./7.)*R_t*r.radius*r.angVel.norm() #intermediate computation
			angVel = (b-a)/r.radius
			r.angVel = angVel * Vt.cross(self.updated_normal)/Vt.norm()
			Vt = Vt/Vt.norm()*(a-b)
			Vn = -R_n*Vn
		else:
			Debug.Print("Sliding bounce")
			angVel = -(5./2.)*(1.+R_n)*Vn.norm()/r.radius + R_t*r.angVel.norm()
			r.angVel = angVel * Vt.cross(self.updated_normal)/Vt.norm()
			Vt = Vt/Vt.norm() * ( R_t*Vt.norm() + np.tan(np.radians(phi))*(1.+R_n)*Vn.norm() )
			Vn = -R_n*Vn
		r.vel=Vn+Vt
		Debug.Print("New vel:",r.vel)
		self.check_output_vel(r,f)


number_to_model_correspondance={0:Classical, 1:Pfeiffer, 2:Bourrier}
"""
Dictionnary that allows to link between the rebound model numbers and the corresponding class.

0: :class:`Classical`
1: :class:`Pfeiffer`
2: :class:`Bourrier`

"""


class Toe_Tree():
	def __init__(self,*args):
		from scipy.spatial import cKDTree
		script_dir = os.path.realpath(__file__).rsplit("/",1)[0]
		vels=[5., 10., 15., 20., 25., 30.]
		self.input_sequence=["vel","vol","ecc","dbh","ang","vx","vy","vz"]
		self.weight_sequence=["vel","vol","ang","ecc","dbh","vx","vy","vz"]	#FIXME: check this sequence
		self.toe_array=np.empty([0,8])	#cols 0->4 are parameters given by input_sequence, cols 5->7 are outputs [vx,vy,vz]
		self.last_computed_data={}
		for vi in range(len(vels)):
			input=np.genfromtxt(script_dir+"/Toe_2018/result"+str(vels[vi])+".txt",skip_header=1)
			input=np.delete(input,[5,6],axis=1) #remove the useless columns (input vels)
			self.toe_array=np.append(self.toe_array,input,axis=0)
		#to sort the array lines we need a "structured view" of it, see https://stackoverflow.com/questions/2828059/sorting-arrays-in-numpy-by-column and https://docs.scipy.org/doc/numpy/user/basics.rec.html
		toe_structured_view=self.toe_array.view({"names":self.input_sequence,'formats':['float']*len(self.input_sequence)})
		toe_structured_view.sort(order=self.weight_sequence, axis=0)
		
	def get_output_vel(self,vel=None,vol=None,ecc=None,dbh=None,ang=None):
		#start with the whole array
		sliced_toe_array=self.toe_array.copy()
		dbh*=0.01 #platrock dbh is in cm, while Toe_2018 is in m.
		for key in self.weight_sequence[:-3]:
			col=self.input_sequence.index(key)	#get the current column index
			real_value=locals()[key]			#get the input value
			id_min=np.argmin(abs(sliced_toe_array[:,col]-real_value))
			id_max=len(sliced_toe_array)-np.argmin(abs(sliced_toe_array[::-1,col]-real_value)) -1
			sliced_toe_array=sliced_toe_array[id_min:id_max+1,:] #narrow the slice at each loop
		Debug.info("Toe model with parameters:",sliced_toe_array[0,0:5])
		self.last_computed_data["output_vel_xyz"]=sliced_toe_array[0,-3:]
		self.last_computed_data["input_vel"]=sliced_toe_array[0,0]
		self.last_computed_data["input_vol"]=sliced_toe_array[0,1]
		self.last_computed_data["input_ecc"]=sliced_toe_array[0,2]
		self.last_computed_data["input_dbh"]=sliced_toe_array[0,3]
		self.last_computed_data["input_ang"]=sliced_toe_array[0,4]
		return self.last_computed_data["output_vel_xyz"]


class Azzoni_Roll():
	#1 - init Azzoni roll, this will compute the stop distance and optionally the stop pos ON AN INFINITE SLOPE
	#2 - Use get_vel(arrival_point) from simulation, regarding the rock environment which this class doesn't know about (if segments ends are reached for example)
	def __init__(self,vel,mu_r,slope,gravity,start_pos=None,tan_slope=None,A=None,mass=None,radius=None,I=None):
		#INPUTS:
			#Mandatory:
		self.vel=vel
		self.mu_r=mu_r
		self.slope=slope
		self.gravity=gravity
			#Optional:
		self.start_pos=start_pos
		if tan_slope is None:
			self.tan_slope=np.tan(slope)
		else:
			self.tan_slope=tan_slope
		if A is None: #give either A or {mass,I,radius}
			self.A = mass/(mass+I/radius**2)
		else: self.A=A
		
		#OUTPUTS:
		self.tan_roll_phi=None
		self.delta_tan_angles=None
		self.dist_stop=None
		self.stop_pos=None #only if start_pos is given
		self.direction=None
		self.v_square=None
		
		#Depending on the rock environment, the variables below can be set by the simulation.
		#OTHER:
		self.until_stop=False
		
		self.compute_dist()
	
	def compute_dist(self):
		self.tan_roll_phi=self.mu_r
		self.direction = -1 if self.vel[0]<0 else 1
		self.tan_slope*=self.direction #if going to -x direction, slope sign convention inverted
		self.delta_tan_angles = self.tan_slope - self.tan_roll_phi
		self.v_square=self.vel.norm()**2
		if(self.delta_tan_angles<0):#slow down rolling
			self.dist_stop=-self.v_square/(2*self.A*self.gravity*np.cos(self.slope)*self.delta_tan_angles)
			if(self.start_pos is not None):
				Dx=np.cos(self.slope)*self.dist_stop*self.direction
				Dy=-self.tan_slope*abs(Dx)
				self.stop_pos=self.start_pos+[Dx,Dy]
		else:
			self.dist_stop=np.inf
			self.stop_pos=Math.Vector2([self.direction*np.inf,self.direction*np.inf])
		
	def get_vel(self,arrival_point):
		roll_dist=(arrival_point-self.start_pos).norm()
		scalar_vel = np.sqrt(2*self.A*self.gravity*np.cos(self.slope)*self.delta_tan_angles*roll_dist+self.v_square)
		return (arrival_point-self.start_pos).normalized()*scalar_vel
		





















"""
Copyright (c) 2017, 2018, 2019 Irstea
Copyright (c) 2017, 2018, 2019 François Kneib
Copyright (c) 2017, 2018, 2019 Franck Bourrier
Copyright (c) 2017, 2018, 2019 David Toe
Copyright (c) 2017, 2018, 2019 Frédéric Berger
Copyright (c) 2017, 2018, 2019 Stéphane Lambert

This file is part of PlatRock.

PlatRock is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, version 3 of the License.

PlatRock is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Foobar. If not, see <http://www.gnu.org/licenses/>.
"""

import platrock
import numpy as np
import platrock.TwoD.Geoms as Geoms
import platrock.Common.Math as Math
import matplotlib as mpl
import plotly.offline as plyo
import plotly.graph_objs as plygo
import platrock.Common.ColorPalettes as cp
import platrock.Common.Outputs as Out

if platrock.web_ui:
	mpl.use('Agg')
	#import plotly as ply
import matplotlib.pyplot as plt
if not platrock.web_ui: plt.ion()

mpl.rcParams["figure.subplot.bottom"]=0.04
mpl.rcParams["figure.subplot.top"]=0.99
mpl.rcParams["figure.subplot.left"]=0.03
mpl.rcParams["figure.subplot.right"]=0.99

fig=plt.figure("PlatRock 2D View",facecolor="w",edgecolor="w")
ax=plt.axes()

def clear():
	global fig,ax
	plt.close("all")
	fig=plt.figure("PlatRock 2D View",facecolor="w",edgecolor="w")
	ax=plt.axes()

def draw_terrain(t):
	seg_points=t.get_points()
	delta_z=max(seg_points[:,1])-min(seg_points[:,1])
	delta_x=max(seg_points[:,0])-min(seg_points[:,0])
	fig.set_size_inches( (18,int(18*1.8*delta_z/delta_x)), forward=True)
	ax.set_ylim(np.min(seg_points[:,1])-0.1*delta_z,np.max(seg_points[:,1])+0.1*delta_z)
	ax.set_xlim(np.min(seg_points[:,0])-0.01*delta_x,np.max(seg_points[:,0])+0.01*delta_x)
	ax.set_aspect("equal")
	ax.fill_between(seg_points[:,0],seg_points[:,1],color="#844200")
	fig.canvas.draw()

def plot_trajectory(s,rock_id,nb_points=10, draw=True, with_contacts=False, use_plotly=False):
	color=s.random_generator.rand(3)
	X=np.array([],dtype=float)
	Y=np.array([],dtype=float)
	pos=s.output.get_contacts_pos(rock_id)
	vels=s.output.get_contacts_vels(rock_id)
	types=s.output.get_contacts_types(rock_id)
	trees_x=[]
	trees_y=[]
	# FIXME : to limit data on TwoDShape simulations:
	if types[1]==Out.MOTION and isinstance(s,platrock.TwoDShape.Simulations.Simulation):
		max_data_points=1000 #limit the nb of points for one rock to 1000
		resample_step=int((len(pos)-1)/max_data_points)+1
		pos=s.output.get_contacts_pos(rock_id)[::resample_step] #NOTE: this is a bit stupid: if we have max_data_points+1 positions, resample_step==2 and data length will be max_data_points/2, but anyway...
	for i in range(len(pos)-1):
		if(types[i] in [Out.START,Out.SOIL,Out.TREE]): #parabola
			x=np.linspace(pos[i][0],pos[i+1][0],nb_points)
			X=np.append(X,x)
			P=Geoms.Parabola(pos=Math.Vector2(pos[i]),vel=Math.Vector2(vels[i]),g=s.gravity)
			Y=np.append(Y,P.A*x*x+P.B*x+P.C)
		elif(types[i]==Out.MOTION):
			X=np.append(X,[pos[i][0]])
			Y=np.append(Y,[pos[i][1]])
		elif(types[i]==Out.ROLL):
			x=[pos[i][0],pos[i+1][0]]
			X=np.append(X,x)
			y=[pos[i][1],pos[i+1][1]]
			Y=np.append(Y,y)
		if(types[i]==Out.TREE or types[i]==Out.ROLL_TREE):
			trees_x.append(pos[i][0])
			trees_y.append(pos[i][1])
	if(use_plotly):
		return X,Y,np.asarray(trees_x),np.asarray(trees_y)
	ax.plot(X,Y,alpha=0.5,lw=2,color=color)
	if(with_contacts):
		normals=s.output.get_contacts_normals(rock_id)
		types=s.output.get_contacts_types(rock_id)
		for i in range(len(normals)):
			if(types[i] in [Out.START,Out.SOIL]):
				vect=Math.Vector2(normals[i]).rotated(-np.pi/2.)
				ax.plot([pos[i][0]-vect[0],pos[i][0]+vect[0]], [pos[i][1]-vect[1],pos[i][1]+vect[1]],"--",color=color)
			elif(types[i] == Out.TREE or types[i]==Out.ROLL_TREE):
				ax.plot(pos[i][0],pos[i][1],"^",ms=7,color=color)
	if(draw):
		fig.canvas.draw()

def plot_trajectories(s,rocks_ids, with_contacts=False):
	for i in rocks_ids:
		plot_trajectory(s,i,draw=False,with_contacts=with_contacts)
	fig.canvas.draw()

def plot_sample_trajectories(s,nb=10, with_contacts=False, use_plotly=False):
	rocks_to_plot=[]
	if(nb>=s.nb_rocks):
		rocks_to_plot=range(s.nb_rocks)
	else:
		rocks_to_plot=np.linspace(0,s.nb_rocks-1,nb,dtype=int)
	if not use_plotly:
		plot_trajectories(s,rocks_to_plot,with_contacts=with_contacts)
	else:
		#this will create [ [X_r1,Y_r1], [X_r2,Y_r2] ...]
		return [ plot_trajectory(s,i,use_plotly=True) for i in rocks_to_plot ]

def plot_checkpoints(s):
	for chkP in s.checkpoints:
		ax.axvline(chkP.x,lw=4,alpha=0.2,color="red")
		#if(hasattr(chkP,"heights")):
			##for i in range(len(chkP.heights)):
				#ax.arrow(chkP.x,chkP.heights[i]+chkP.base_height,chkP.vels[i][0],chkP.vels[i][1],head_width=2,head_length=2)
	fig.canvas.draw()

def get_plotly_raw_html(s, sample_trajectories=False):
	#Initialize the lists
	data = []
	annotations = []
	#1- plot the terrain:
	seg_points = s.terrain.get_points()
	terrain_dx=seg_points[:,0].max()-seg_points[:,0].min()
	terrain_dy=seg_points[:,1].max()-seg_points[:,1].min()
	bbox=[	#bounding box: [ minx, maxx, miny, maxy ]
		seg_points[:,0].min()-0.05*terrain_dx,
		seg_points[:,0].max()+0.05*terrain_dx,
		seg_points[:,1].min()-0.05*terrain_dy,
		seg_points[:,1].max()+0.05*terrain_dy,
	]
	
	i=0
	zones_polygon_centers=[]
	zones_polygon_colors=[]
	for zone in s.terrain.params_zones:
		zone_pts=seg_points[zone.start_id:zone.end_id+2,:]
		zones_polygon_centers.append(Math.get_2D_polygon_center_of_mass(np.append(zone_pts,[[zone_pts[-1,0],bbox[2]],[zone_pts[0,0],bbox[2]]],axis=0)))
		zones_polygon_colors.append(cp.color_to_html(zone["color"]))
		zero_line = plygo.Scatter(
			x=[zone_pts[0,0],zone_pts[-1,0]],
			y=[bbox[2]]*2,
			fill='none',
			line=dict(
					color="rgba(255,255,255,0)",
					width=0
			),
			hoverinfo='skip',
			mode="lines",
			name='zones'
		)
		terrain = plygo.Scatter(
			x=zone_pts[:,0],
			y=zone_pts[:,1],
			mode='lines+markers',
			fill='tonexty',
			text="Point",
			line=dict(
				color=zones_polygon_colors[-1],
				width=3
			),
			name='terrain',
			hoverinfo='text+x+y',
		)
		data.append(zero_line)
		data.append(terrain)
		i+=1
		
	#1.1- plot markers at zones centers to handle hoverinfo/clicks events
	zones_polygon_centers=np.asarray(zones_polygon_centers)
	data.append(plygo.Scatter(
		x=zones_polygon_centers[:,0],
		y=zones_polygon_centers[:,1],
		text = ["Zone "+str(i) for i in range(len(zones_polygon_centers))],
		mode = 'markers',
		marker = dict(
			symbol = 'circle',
			size = 15,
			color = zones_polygon_colors,
		),
		hoverinfo='text',
		name="zones"
	))
	
	#2- plot the checkpoints:
	for chkP in s.checkpoints:
		data.append(
			plygo.Scatter(
				x=[chkP.x]*2,
				y=[s.terrain.get_y(chkP.x), 1e20],
				mode = 'lines',
				line=dict(
					color="rgba(255,0,0,0.5)",
					width=4
				),
				hoverinfo='skip', #dont show datapoints on mouse over
				name='',
			)
		)
		if(sample_trajectories and len(chkP.heights)>0):
			annotations.append(
				dict(
					x=chkP.x,
					y=1,
					xref="x",
					yref="paper",
					text=str(round(chkP.crossings_ratio*100,1))+"%",
					showarrow=False,
					textangle=-90,
					bordercolor="rgba(255,0,0,0.5)",
					borderwidth=4,
					borderpad=1,
					bgcolor="rgba(255,255,255,1)",
					hovertext="Mean height: "+str(round((np.asarray(chkP.heights).mean()),2))+"<br>Mean vel: "+str(round((np.linalg.norm(np.asarray(chkP.vels),axis=1).mean()),2))
			)
		)
	
	#3- plot sample trajectories:
	if(sample_trajectories):
		rocks_data = plot_sample_trajectories(s,nb=sample_trajectories, use_plotly=True)
		for rock_data in rocks_data:
			color=s.random_generator.randint(255,size=3)
			color="rgb("+",".join(color.astype(str))+")"
			#parabolas:
			data.append(
				plygo.Scatter(
					x=rock_data[0],
					y=rock_data[1],
					mode = 'lines',
					line=dict(
						color=color,
						width=1
					),
					hoverinfo='skip', #dont show datapoints on mouse over
					name='',
				)
			)
			#trees:
			data.append(
				plygo.Scatter(
					x=rock_data[2],
					y=rock_data[3],
					mode = 'markers',
					marker = dict(
						symbol = 'triangle-up',
						size = 6,
						#line = dict(width=2),
						color = color,
						opacity = 1
					),
					hoverinfo='skip', #dont show datapoints on mouse over
					name='',
				)
			)
		cdf=s.get_stops_cdf()
		data+=[
			plygo.Scatter(
				x=[seg_points[0,0],seg_points[-1,0]],
				y=[bbox[2]]*2,
				fill='none',
				line=dict(
						color="rgba(255,255,255,0)",
						width=0
				),
				hoverinfo='skip',
			),
			plygo.Scatter(
				x=cdf[0,:],
				y=cdf[1,:]*(s.terrain.get_y_range()[1]-bbox[2])+bbox[2],
				fill='tonexty',
				text=["CDF = "+str(value) for value in cdf[1,:]],
				name="CDF",
				hoverinfo='text',
				mode = 'lines',
				line=dict(
					color="rgba(0,0,0,0.7)",
					width=2,
					shape='hv',
				),
				fillcolor="rgba(0,0,0,0.15)",
			)
		]
	
	#4- Configure layout (axes, etc...):
	xrange=bbox[:2]
	yrange=bbox[2:]
	layout = plygo.Layout(
		autosize=False,
		width=1000,
		height=800,
		margin=plygo.layout.Margin(
			l=40,
			r=10,
			b=20,
			t=30,
			pad=4
		),
		xaxis = dict(
			#nticks = 10,
			linecolor= 'black',
			mirror=True,
			showgrid=True,
			range = xrange,
			zeroline=False,
		),
		yaxis = dict(
			linecolor= 'black',
			mirror=True,
			showgrid=True,
			zeroline=False,
			scaleanchor = "x", #to make 1:1 scale
			range = yrange,
		),
		showlegend= False,
		hovermode = 'closest',
		annotations=annotations,
	)
	
	#5- configure plotly menubar:
	config = dict(
		modeBarButtonsToRemove = ["autoScale2d","toggleSpikelines","hoverCompareCartesian","hoverClosestCartesian","select2d","lasso2d"],
		displaylogo = False,
		displayModeBar = True,
	)
	
	#6- Generate plotly html/js:
	ply_fig = plygo.Figure(data=data, layout=layout)
	return plyo.plot(ply_fig, config=config, show_link=False, output_type="div", include_plotlyjs=False)

def get_plotly_rock_raw_html(s,r):
	data = []
	
	data.append(
		plygo.Scatter(
			x=[0],
			y=[0],
			text="COG",
			hoverinfo="text",
			marker = dict(
				symbol = 'circle',
				size = 10,
				color = "black"
			),
		)
	)
	
	points=np.copy(r.vertices)
	if not s.random_rocks_ori:
		Math.rotate_points_around_origin(points,np.radians(s.rocks_ori))
	vertices = plygo.Scatter(
		x=points[:,0].tolist()+[points[0,0]],
		y=points[:,1].tolist()+[points[0,1]],
		hoverinfo="x+y",
		line=dict(
			color="rgb(102, 102, 102)",
			width=3
		),
	)
	data.append(vertices)
	shapes=[
		plygo.layout.Shape(
				type="path",
				path='M'+"L".join([','.join(i) for i in points.astype(str)])+'Z',
				fillcolor='rgba(102, 102, 102, 0.62)',
				line=dict(
					#color=zones_polygon_colors[-1],
					width=0
				),
		)
	]

	layout = plygo.Layout(
		width=400,
		height=400,
		margin=plygo.layout.Margin(
			l=40,
			r=10,
			b=20,
			t=30,
			pad=4
		),
		xaxis = dict(
			mirror=True,
			showgrid=True,
			linecolor="black",
			linewidth=1,
		),
		yaxis = dict(
			mirror=True,
			showgrid=True,
			scaleanchor = "x", #to make 1:1 scale
			linecolor="black",
			linewidth=1,
		),
		showlegend= False,
		hovermode = 'closest',
		shapes=shapes,
	)

	config = dict(
		displayModeBar= False,
		responsive= True,
	)
	
	ply_fig = plygo.Figure(data=data, layout=layout)
	return plyo.plot(ply_fig, config=config, show_link=False, output_type="div", include_plotlyjs=False)














"""
Copyright (c) 2017, 2018, 2019 Irstea
Copyright (c) 2017, 2018, 2019 François Kneib
Copyright (c) 2017, 2018, 2019 Franck Bourrier
Copyright (c) 2017, 2018, 2019 David Toe
Copyright (c) 2017, 2018, 2019 Frédéric Berger
Copyright (c) 2017, 2018, 2019 Stéphane Lambert

This file is part of PlatRock.

PlatRock is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, version 3 of the License.

PlatRock is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Foobar. If not, see <http://www.gnu.org/licenses/>.
"""

"""
This module is used by the ThreeD model. It handles all the Objects types
"""

import numpy as np
import platrock.Common.Debug as Debug
import platrock
import quaternion
import platrock.Common.Math as Math
import copy,sys
from . import RasterTools
import osgeo.gdal as gdal
import shapely.geometry
import shapely.affinity
import platrock.Common.ColorPalettes as cp

class Point(object):
	"""
	A point, which is composing the terrain faces vertex or the rock vertex.
	
	Args:
		id (int): a unique identifier used for the 3D view
		pos (:class:`Common.Math.Vector3` [float,float,float]): absolute x,y,z position components
		vel (:class:`Common.Math.Vector3` [float,float,float]): absolute vx,vy,vz velocity components
		relPos (:class:`Common.Math.Vector3` [float,float,float]): if the point is member of a :class:`Triangulated_Object`, the refPos is the position of the point in the object coordinates (centered at COG).
		VTK_actor (:class:`vtk.VtkActor`): for 3D view
		color (list [1.,0.,0.]): the color of the object for raster or 3D views
		radius (float): the point radius, if any
		TO (:class:`Triangulated_Object`): the triangulated_object the point is member of
		faces (list [:class:`Face`]): the list of all faces the point is member of
	"""
	def __init__(self,x=0,y=0,z=0,TO=None):
		"""
		Constructor:
		
		Args:
			x(float): the x in pos
			y(float): the y in pos
			z(float): the z in pos
		"""
		Triangulated_Object.points_counter+=1
		self.id=Triangulated_Object.points_counter
		self.pos=Math.Vector3([x,y,z])
		self.vel=Math.Vector3([0.,0.,0.])
		self.relPos=Math.Vector3([x,y,z])
		self.VTK_actor=None
		self.color=[1.,0.,0.]
		self.radius=0.1
		self.TO=TO
	
	def update_pos(self):
		"""
		Update the position of the point according to its TO attibute.
		"""
		self.pos=quaternion.rotate_vector(self.TO.ori,self.relPos)
		self.pos+=self.TO.pos

class Face(object):
	"""
	A face, composing a rock or the terrain.
	
	Args:
		id (int): a unique identifier used for the 3D view
		points (:class:`numpy.ndarray` [:class:`Point`,:class:`Point`,:class:`Point`]): an array of 3 points linked by the face
		normal (:class:`Common.Math.Vector3` [float, float, float]): the normal vector of the face
		color (list [1.,0.,0.]): the color of the object for raster or 3D views
		VTK_actor (:class:`vtk.VtkActor`): for 3D view
		TO (:class:`Triangulated_Object`): the triangulated_object the face is member of
		roughness (float): the roughness of the terrain for this face
	"""
	def __init__(self,point1,point2,point3,bounce_model_number=None,R_t=None,R_n=None,roughness=None,v_half=None,phi=None,TO=None,**kwargs):	#allow **kwargs here as e and mu (siconos) parameters will be given but not used here
		"""
		Args:
			point1 (:class:`Point`): the first point the face is connected to
			point2 (:class:`Point`): the second point the face is connected to
			point3 (:class:`Point`): the third point the face is connected to
		FIXME : DOC
		"""
		#Triangulated_Object.faces_counter+=1
		#self.id=Triangulated_Object.faces_counter
		self.points=[point1,point2,point3]
		self.normal=None
		self.cog=None
		self.color=[0.68235294, 0.5372549 , 0.39215686]
		self.VTK_actor=None
		self.TO=TO
		self.xyz_bb=None
		#used only if the face belongs to the terrain:
		self.roughness=roughness
		self.R_n=R_n
		self.R_t=R_t
		self.bounce_model_number=bounce_model_number
		self.v_half=v_half
		self.phi=phi
		self.xyz_bb=np.array([	min(self.points[0].pos[0],self.points[1].pos[0],self.points[2].pos[0]),
								max(self.points[0].pos[0],self.points[1].pos[0],self.points[2].pos[0]),
								min(self.points[0].pos[1],self.points[1].pos[1],self.points[2].pos[1]),
								max(self.points[0].pos[1],self.points[1].pos[1],self.points[2].pos[1]),
								min(self.points[0].pos[2],self.points[1].pos[2],self.points[2].pos[2]),
								max(self.points[0].pos[2],self.points[1].pos[2],self.points[2].pos[2])])
	
	def is_point_inside_2D(self,point):
		"""
		From an arbitrary point
		"""
		point_xy=point.pos[0:2]	#
		face_points_xy=[p.pos[0:2] for p in self.points ]
		cross_products=[Math.cross2(face_points_xy[i-2]-face_points_xy[i],point_xy-face_points_xy[i]) for i in range(0,3)]
		cross_products_signs_sum=sum(np.sign(cross_products))
		return abs(cross_products_signs_sum)==3
	
	def is_point_inside_3D(self,point):
		"""
		This method is not used anymore. Better do this on 2D ("view from the sky") as it avoids convex edge issues (non detection by the contacts_detector).
		"""
		point_on_face=point.pos - (point.pos-self.points[0].pos).dot(self.normal)*self.normal
		cp1=((self.points[-2].pos-self.points[0].pos).cross(point_on_face-self.points[0].pos)).dot(self.normal)
		cp2=((self.points[-1].pos-self.points[1].pos).cross(point_on_face-self.points[1].pos)).dot(self.normal)
		cp3=((self.points[0].pos-self.points[2].pos).cross(point_on_face-self.points[2].pos)).dot(self.normal)
		#cross_products=[(self.points[i-2].pos-self.points[i].pos).cross(point_on_face-self.points[i].pos) for i in range(0,3)]
		#dots=[cp.dot(self.normal) for cp in cross_products]
		cross_products_signs_sum=np.sign(cp1)+np.sign(cp2)+np.sign(cp3)
		return abs(cross_products_signs_sum)==3
		
			

# /!\ ALL TRIANGULATED OBJECTS MUST INHERIT FROM THIS CLASS !
class Triangulated_Object(object):
	def __init__(self):
		self.points=np.array([])
		self.faces=np.array([])
		self.VTK_actor=None	#VTK actor for visualization
		self.opacity=1
		self.enable_points_view=False	
		self.points_as_array=np.array([])
	
	def set_points_as_array(self):
		self.points_as_array=np.asarray([p.relPos for p in self.points])
	points_counter=0


class Rock(Triangulated_Object):
	"""A falling rock."""
	def __init__(self,pos=[0.,0.,0.],vel=[0.,0.,0.],angVel=[0.,0.,0.],density=2700.):
		Triangulated_Object.__init__(self)
		self.pos = Math.Vector3(pos)
		self.ori=quaternion.from_euler_angles(0,0,0)
		self.vel=Math.Vector3(vel)
		self.volume=None
		self.mass=None
		self.I=None
		self.angVel=Math.Vector3(angVel)
		self.density=density
		self.is_stopped=False
		self.color=np.random.rand(3).tolist()
		self.contacts=[]
		self.force=Math.Vector3([0.,0.,0.])
		self.torque=Math.Vector3([0.,0.,0.])
		self.angMom=Math.Vector3([0.,0.,0.])
		self.verlet_faces_list=[]
		self.verlet_trees_list=[]
		self.terrain_active_contact=None
		self.tree_active_contact=None
		self.out_of_bounds=False
		
	def init_mass_inertia(self):
		self.volume=0.
		points=np.asarray([p.pos for p in self.points])
		point_inside=np.mean(points,axis=0)
		G=Math.Vector3()
		for f in self.faces:
			A,B,C=f.points[0].pos,f.points[1].pos,f.points[2].pos
			face_surface = ((B-A).cross(C-A)).norm()/2.
			face_normal = ((A-B).cross(C-B)).normalized()
			if(face_normal.dot(point_inside-A)>0.):
				face_normal*=-1.
			tetrahedron_volume=np.abs(face_surface*face_normal.dot(point_inside-A)/3.)
			tetrahedron_G=(A+B+C+point_inside)/4.
			G+=tetrahedron_volume*tetrahedron_G
			self.volume+=tetrahedron_volume
			
		self.mass=self.density*self.volume
		self.radius=(3.*self.volume/4./np.pi)**(1./3.)
		G/=self.volume
		for p in self.points:
			p.relPos-=G			
		self.I=2./5.*self.mass*self.radius**2
		self.I=Math.Vector3([self.I,self.I,self.I])
		self.update_members_pos()
	
	def update_members_pos(self):
		for p in self.points:
			p.update_pos()


class Sphere(Rock):
	"""
	FIXME: DOC
	"""
	def __init__(self,volume=1,**kwargs):
		super(Sphere,self).__init__(**kwargs)
		self.volume=volume
		self.radius=(self.volume*3/4/np.pi)**(1/3)
		self.points=np.array([Point(self.pos[0],self.pos[1],self.pos[2],TO=self)])	#a sphere has a single point at the center
		self.points[0].pos=self.pos #this is a "pointer", all updates to self.pos will update self.points[0]
		self.points[0].radius=self.radius #this will only be used for visualization
		self.enable_points_view=False
		self.bounding_box=Bounding_Box()
		self.init_mass_inertia()
		
	def init_mass_inertia(self):
		self.mass=self.density*self.volume
		self.I=np.ones(3)*2./5.*self.mass*self.radius**2
	def update_members_pos(self):
		pass
		

class Parallelepiped(Rock):
	def __init__(self,lengths=[1,1,1],**kwargs):
		self.lengths=Math.Vector3(lengths)
		super(Parallelepiped,self).__init__(**kwargs)
		for x in [-lengths[0]/2.,+lengths[0]/2.]:
			for y in [-lengths[1]/2.,+lengths[1]/2.]:
				for z in [-lengths[2]/2.,+lengths[2]/2.]:
					self.points=np.append(self.points,Point(x,y,z,TO=self) )
		self.faces=[
			Face( self.points[0], self.points[1], self.points[2], TO=self ),
			Face( self.points[1], self.points[2], self.points[3], TO=self ),
			Face( self.points[0], self.points[1], self.points[5], TO=self ),
			Face( self.points[0], self.points[4], self.points[5], TO=self ),
			Face( self.points[1], self.points[5], self.points[7], TO=self ),
			Face( self.points[1], self.points[3], self.points[7], TO=self ),
			Face( self.points[2], self.points[3], self.points[6], TO=self ),
			Face( self.points[3], self.points[6], self.points[7], TO=self ),
			Face( self.points[0], self.points[2], self.points[4], TO=self ),
			Face( self.points[2], self.points[4], self.points[6], TO=self ),
			Face( self.points[4], self.points[5], self.points[6], TO=self ),
			Face( self.points[5], self.points[6], self.points[7], TO=self ),
		]
		self.init_mass_inertia()
		self.set_points_as_array()
	
	def get_siconos_convex_hull(self):
		assert(len(self.points)>3)
		from siconos.mechanics.collision import SiconosConvexHull, SiconosBox
		#hull = SiconosConvexHull(self.points_as_array)
		hull = SiconosBox(self.lengths)
		dims = [self.points_as_array[:,0].max() - self.points_as_array[:,0].min(),
				self.points_as_array[:,1].max() - self.points_as_array[:,1].min(),
				self.points_as_array[:,2].max() - self.points_as_array[:,2].min()]
		hull.setInsideMargin(min(dims)*0.02)
		hull.setOutsideMargin(0)
		return hull

from . import ThreeDEnginesToolbox
class Terrain(Triangulated_Object):
	valid_input_geojson_attrs=np.array([
		["R_n",					"R<sub>n</sub>",		float,	0,	0.95	],
		["R_t",					"R<sub>t</sub>",		float,	0,	0.95	],
		["roughness",			"Roughness",			float,	0,	1.		],
		["bounce_model_number",	"Rebound model",		int,	0,	2		],
		["phi",					"φ",					float,	0,	40		],
		["v_half",				"V<sub>1/2</sub>",		float,	0,	50		],
		["e",					"e",					float,	0,	0.95	],
		["mu",					"μ",					float,	0,	0.95	],
		["mu_r",				"μ<sub>r</sub>",		float,	0,	100		],
	])
	def __init__(self,DEM_asc=None,params_geojson=None,default_faces_params={"R_n":0.6, "R_t":0.6, "roughness":0.1,"bounce_model_number":0,"phi":30,"v_half":2.5,"e":0.1,"mu":0.8,"mu_r":0.5}):
		Triangulated_Object.__init__(self)
		self.enable_points_view=False
		self.color=[0.68235294, 0.5372549 , 0.39215686]
		self.trees=[]
		self.trees_as_array=None
		self.vertical_surface=0.
		self.Z_raster=None
		self.opacity=1
		self.faces_xyz_bb=None
		self.soil_params=[]
		self.default_faces_params=default_faces_params
		self._geojson_polygon_offset_applied=False
		if(DEM_asc is not None):
			if(platrock.web_ui):	#don't populate/precompute here, do it in s.before_run_tasks()
				self.set_Z_raster_from_DEM_file(DEM_asc)
			else:
				self.from_DEM_file(DEM_asc)
		if(params_geojson is not None): #NOTE: the function below needs the Z_raster to be set, as it needs the xllcorner/yllcorner data
			self.set_soil_params_from_geojson(params_geojson)

	def from_DEM_file(self,filename):
		self.set_Z_raster_from_DEM_file(filename)
		self.populate_from_Z_raster()
		self.precompute_datas()
	
	def precompute_datas(self):
		self.set_faces_normals_and_cogs()
		self.set_points_as_array()
		self.set_faces_xyz_bbs()
		self.min_height=self.points_as_array[:,2].min()
		
	def set_Z_raster_from_DEM_file(self,filename):
		try:
			self.Z_raster=RasterTools.from_asc(filename,"Z")
			return 0
		except Exception as e:
			message="Failed to import the Esri grid file:"+str(e)
			Debug.error(message)
			return message
	
	def set_soil_params_from_geojson(self,params_geojson):
		try:
			self._geojson_polygon_offset_applied=False
			shp_file=gdal.OpenEx(params_geojson)
			self.soil_params=[]
			for feat in shp_file.GetLayer(0):
				feat_dict=feat.ExportToJson(as_object=True)
				self.soil_params.append({})
				shapely_poly = shapely.geometry.shape(feat_dict['geometry'])
				self.soil_params[-1]["shapely_polygon"]=shapely_poly
				self.soil_params[-1]["color"]=cp.qualitative10[len(self.soil_params)-1]
				soil_params = feat_dict['properties']
				#geojson properties names has limited char count, so replace :
				if("bounce_mod" in list(soil_params.keys())):
					soil_params["bounce_model_number"] = soil_params.pop("bounce_mod")
				#validate the parameters from the input geojson
				for k in list(soil_params.keys()):
					if(k not in Terrain.valid_input_geojson_attrs[:,0]):
						soil_params.pop(k)
				self.soil_params[-1]["params"]=soil_params
			return 0
		except Exception as e:
			message="Failed to import the geojson:"+str(e)
			Debug.error(message)
			return message
		
	def set_faces_params_from_geojson(self):
		#loop on faces to set attrbutes
		for face in self.faces:	#FIXME: this loop can be slow, optimize by replacing shapely "contains" by a cython or numpy array algo ?
			shapely_point=shapely.geometry.Point(face.cog[0:2])
			for soil_param in self.soil_params:
					if(soil_param["shapely_polygon"].contains(shapely_point)):
						face.__dict__.update(soil_param["params"])	#update will overwrite params
						break
	
	def populate_from_Z_raster(self):
			points_ids=np.ones([self.Z_raster.nx,self.Z_raster.ny])*-1 #this is a map of points ids, useful to link [x,y] cells to corresponding point ids.
			#CREATE POINTS :
			nb_points=np.count_nonzero(self.Z_raster.data["Z"]!=self.Z_raster.header_data["NODATA_value"])
			self.points=np.empty(nb_points,dtype=Point)	#to avoid increasing the size of the points array at each new point, instanciate here with the right size.
			counter=0
			for i in range(0,self.Z_raster.nx):
				for j in range(0,self.Z_raster.ny):
					if(not self.Z_raster.data["Z"][i,j]==self.Z_raster.header_data["NODATA_value"]):
						self.points[counter]=Point(self.Z_raster.X[i],self.Z_raster.Y[j],self.Z_raster.data["Z"][i,j])
						points_ids[i,j]=counter
						counter+=1
			# CREATE TRIANGLES :
			# id0---id1
			#  |   / |
			#  |  /  |
			#  | /   |
			# id2---id3
			#to avoid increasing the size of the faces array at each new face, instanciate here.
			#We don't know the exact size now, so use 2*nb_points wich is the worst-case, then finally decrease the size.
			self.faces=np.empty(2*nb_points,dtype=Face)
			counter=0
			for i in range(0,len(self.Z_raster.X)-1):
				for j in range(0,len(self.Z_raster.Y)-1):
					id0=int(points_ids[i,j])
					if(id0!=-1):
						id1=int(points_ids[ i+1 , j   ])
						id2=int(points_ids[ i   , j+1 ])
						id3=int(points_ids[ i+1 , j+1 ])
						if(id1!=-1 and id2!=-1):
							self.faces[counter]=Face(self.points[id0], self.points[id1], self.points[id2],**self.default_faces_params)
							counter+=1
							if(id3!=-1):
								self.faces[counter]=Face( self.points[id1], self.points[id2], self.points[id3],**self.default_faces_params)
								counter+=1
			self.faces=self.faces[0:counter]#clean the array
	
	
	def set_faces_normals_and_cogs(self):
		ThreeDEnginesToolbox.set_faces_normals_and_cogs(self.faces)
		#for f in self.faces:
			#normal=((f.points[0].pos-f.points[1].pos).cross(f.points[2].pos-f.points[1].pos)).normalized()
			#if(normal[2]<0.):normal*=-1.
			#f.normal=normal
			#f.cog=(f.points[0].pos + f.points[1].pos + f.points[2].pos)/3
	
	def set_faces_xyz_bbs(self):
		self.faces_xyz_bb=np.asarray([f.xyz_bb for f in self.faces])
			
	def set_vertical_surface(self):
		#compute vertical surface (=surface seen from the sky) :
		self.vertical_surface=0
		for f in self.faces:
			self.vertical_surface += 0.5*np.abs(Math.cross2(f.points[1].pos[0:2]-f.points[0].pos[0:2],f.points[2].pos[0:2]-f.points[0].pos[0:2]))
			
	def generate_random_forest(self,density,dhp_mean, dhp_std):
		self.set_vertical_surface()
		nb_trees=int(density*self.vertical_surface/10000) #NOTE: density is in trees/ha, vertical_surface is in m^2
		nb_faces=len(self.faces)
		self.trees=[]
		for i in range(nb_trees):
			pick_face = self.faces[int(np.random.random()*nb_faces)]
			pt1_xy=pick_face.points[0].pos[0:2]
			pt2_xy=pick_face.points[1].pos[0:2]
			pt3_xy=pick_face.points[2].pos[0:2]
			rand_point = pt1_xy+np.random.random()*(pt2_xy-pt1_xy) # pick a point somewhere between pt1 and pt2
			rand_point = rand_point + np.random.random()*(pt3_xy-rand_point)	# move the previous picked point toward pt3
			dhp = Math.get_random_value_from_gamma_distribution(dhp_mean,dhp_std)
			self.trees.append(Tree(pos=rand_point,dhp=dhp))
	
	def get_vtk_string(self): #NOTE: deprecated
		import vtk
		points = vtk.vtkPoints()
		triangles = vtk.vtkCellArray()
		
		for p in self.points:
			points.InsertPoint(p.id,p.relPos[0],p.relPos[1],p.relPos[2])
			
		for f in self.faces:
			triangle = vtk.vtkTriangle()
			triangle.GetPointIds().SetId(0,f.points[0].id)
			triangle.GetPointIds().SetId(1,f.points[1].id)
			triangle.GetPointIds().SetId(2,f.points[2].id)
			triangles.InsertNextCell(triangle)
			
		polydata = vtk.vtkPolyData()
		polydata.SetPoints( points )
		polydata.SetPolys( triangles )
		
		writer = vtk.vtkXMLPolyDataWriter()
		writer.SetWriteToOutputString(1)
		if vtk.VTK_MAJOR_VERSION <= 5:
			writer.SetInput(polydata)
		else:
			writer.SetInputData(polydata)
		writer.Update()
		return writer.GetOutputString()
	
	def get_siconos_mesh(self):
		from siconos.mechanics.collision import SiconosMesh
		points=self.points_as_array.transpose()
		triangles=np.empty(len(self.faces)*3, dtype=np.int)
		for fi in range(len(self.faces)):
			triangles[fi*3+0]=self.faces[fi].points[0].id-1		# "-1" as id numbering starts at 1 while here it sould start at 0
			triangles[fi*3+1]=self.faces[fi].points[1].id-1
			triangles[fi*3+2]=self.faces[fi].points[2].id-1
		shape = SiconosMesh(list(triangles), points)
		dims = points.max(axis=0) - triangles.min(axis=0)
		shape.setInsideMargin(min(dims)*0.02)
		shape.setOutsideMargin(0)
		return shape
	
	def get_siconos_heightmap(self):
		from siconos.mechanics.collision import SiconosHeightMap
		z=self.Z_raster.data["Z"][:,:]
		DX=self.Z_raster.X[-1]
		DY=self.Z_raster.Y[-1]
		offsets=[DX/2,DY/2]	#because SiconosHeightmap is a rectangle heightmap centered at zero.
		return SiconosHeightMap(z,DX,DY),offsets

class Contact(object):
	def __init__(self):
		self.dist=None
		self.rock_force=None
		self.rock_torque=None
		self.rock_pos=None
		self.rock_output_vel=None
		self.rock_output_angVel=None
		self.rock_output_ori=None
	
	def get_storage_copy(self,r):
		c=copy.copy(self)
		c.rock_force=r.force.copy()
		c.rock_torque=r.torque.copy()
		c.rock_pos=r.pos.copy()
		c.rock_output_velocity=r.vel.copy()
		c.rock_output_angVel=r.angVel.copy()
		c.rock_output_ori=copy.copy(r.ori)
		return c

class Rock_terrain_contact(Contact):
	def __init__(self,point,face):
		super(Rock_terrain_contact,self).__init__()
		self.point=point	#this is the rock point, not the contact point.
		self.face=face
	
	def get_storage_copy(self, r):
		c=super(Rock_terrain_contact,self).get_storage_copy(r)
		#NOTE : the objects below are not copied, only linked
		c.point=self.point
		c.face=self.face
		return c
		

class Rock_tree_contact(Contact):
	def __init__(self,tree):
		super(Rock_tree_contact,self).__init__()
		self.tree=tree
	
	def get_storage_copy(self, r):
		c=super(Rock_tree_contact,self).get_storage_copy(r)
		c.tree_pos=self.tree.pos.copy()
		#NOTE : the object below is not copied, only linked
		c.tree=self.tree
		return c

class Tree(object):
	def __init__(self,pos,dhp=30):
		self.pos = Math.Vector2(pos) #x,y
		self.dhp = dhp
		self.active = True
		self.color = [0,0.8,0]
		self.VTK_actor=None	#VTK actor for visualization

class Bounding_Box(object):
	def __init__(self):
		self.pos=Math.Vector3([0,0,0])
		self.half_length=0	#set this to 0 so that the Verlet algorithm detect that the bounding sphere has not yet been initialized.
		self.VTK_actor=None
		self.opacity=0.25

class Checkpoint(object):
	"""
	A checkpoint along the terrain. Checkpoints data are post-computed from all contacts of all rocks.
	
	Args:
		x (float): the x-position of the checkpoint
		rocks (list [:class:`Rock`, ...]): all rocks that passed through the checkpoint
		heights (list [float, ...]): all heights at which the rocks passed through the checkpoint
		vels (list [[vx1,vz1], [vx2,vz2], ...]): all velocities at which the rocks passed through the checkpoint
		angVels (list [float, ...]): all angVels at which the rocks passed through the checkpoint
	"""
	def __init__(self,path):
		self.path=np.asarray(path)

	def init_data(self,simulation):
		"""
		Initialize data lists: :attr:`rocks`, :attr:`heights`, :attr:`vels`, :attr:`angVels`
		"""
		self.pos=[]
		self.vels=[]
		self.angVels=[]
		self.rocks_ids=[]
		self.shapely_linestring=shapely.geometry.linestring.asLineString(self.path-[simulation.terrain.Z_raster.xllcorner,simulation.terrain.Z_raster.yllcorner])







from osgeo import gdal
import functools,operator



def get_geojson_all_properties(filename,unique=True):
	result=[]
	shp_file=gdal.OpenEx(filename)
	for feat in shp_file.GetLayer(0):
		feat_dict=feat.ExportToJson(as_object=True) #loop on rocks start polygons
		result.append(list(feat_dict['properties'].keys()))
	if unique:
		result=functools.reduce(operator.iconcat, result, []) #flatten
		unique_result = list(set(result))
		return unique_result
	return result

def get_geojson_all_shapes(filename,unique=True):
	result=[]
	shp_file=gdal.OpenEx(filename)
	for feat in shp_file.GetLayer(0):
		feat_dict=feat.ExportToJson(as_object=True) #loop on rocks start polygons
		result.append(feat_dict['geometry']["type"])
	if unique:
		#result=functools.reduce(operator.iconcat, result, []) #flatten
		unique_result = list(set(result))
		return unique_result
	return result



"""
Copyright (c) 2017, 2018, 2019 Irstea
Copyright (c) 2017, 2018, 2019 François Kneib
Copyright (c) 2017, 2018, 2019 Franck Bourrier
Copyright (c) 2017, 2018, 2019 David Toe
Copyright (c) 2017, 2018, 2019 Frédéric Berger
Copyright (c) 2017, 2018, 2019 Stéphane Lambert

This file is part of PlatRock.

PlatRock is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, version 3 of the License.

PlatRock is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Foobar. If not, see <http://www.gnu.org/licenses/>.
"""

"""
This module is used to store output data in a optimized hdf5 file.
"""

import h5py
import numpy as np
import platrock.Common.Debug as Debug

float_type = np.float32

#Contact types :
START=0
SOIL=1
TREE=2
ROLL_TREE=3	#contact with a tree while rolling
ROLL=4		#for rolling-friction algo
STOP=5		#at the end if the rock was stopped
OUT=6		#at the end if the rock went outside the terrain
FLY=7		#used for ThreeD.Postprocessings to describe a fly-over-raster-cell
MOTION=8	#used by TwoDShape to track the motion of the rocks (don't really track contacts)

class Contact_field(object):
	def __init__(self,name,slice):
		self.name=name
		self.slice=slice

class Output(object):
	def __init__(self,s):
		self._s=s
		
		self._rocks_counter=0
		sim_type=self._s.__module__
		if sim_type=="platrock.TwoD.Simulations" or sim_type=="platrock.TwoDShape.Simulations" :
			self._ndim=2
			self.contacts_slices = {
				"types":slice(0,1),
				"rock_pos":slice(1,3),
				"rock_output_vels":slice(3,5),
				"rock_output_angVels":slice(5,6),
				"normals":slice(6,8)
			}
		elif sim_type=="platrock.ThreeD.Simulations" :
			self._ndim=3
			self.contacts_slices = {
				"types":slice(0,1),
				"rock_pos":slice(1,4),
				"rock_output_vels":slice(4,7),
				"rock_output_angVels":slice(7,10),
				"normals":slice(10,13)
			}
		self._contacts_fields_length=0
		for cs in self.contacts_slices.values():
			self._contacts_fields_length+=cs.stop-cs.start
		
		self.checkpoints=None #this is a list directly exported from the "update_checkpoints" function
		self.volumes=np.empty([self._s.nb_rocks],dtype=float_type)
		self.densities=np.empty([self._s.nb_rocks],dtype=float_type)
		if self._ndim==2:
			self.inertias=np.empty([self._s.nb_rocks],dtype=float_type)
		else:
			self.inertias=np.empty([self._s.nb_rocks,3,3],dtype=float_type) #3X3 matrix
		self.contacts=[]

	def add_rock(self,r):
		self.volumes[self._rocks_counter]=float_type(r.volume)
		self.densities[self._rocks_counter]=float_type(r.density)
		if self._ndim==2:
			self.inertias[self._rocks_counter]=float_type(r.I)
		elif r.I.shape==(3,): #diagonal
			self.inertias[self._rocks_counter]=float_type(np.identity(3)*r.I)
		elif r.I.shape==(3,3): #matrix
			self.inertias[self._rocks_counter]=r.I
		else:
			Debug.error("This rock has an invalid inertia shape:",r.I)
		self.contacts.append(np.empty([0,self._contacts_fields_length],dtype=float_type))
		self._rocks_counter+=1
	
	def add_contact(self,r,normal,type_):
		new_contact_data=np.concatenate((
			[type_],
			r.pos,
			r.vel,
			r.angVel,
			normal
		)).astype(float_type)
		self.contacts[-1]=np.append(self.contacts[-1],[new_contact_data],axis=0)
	
	def del_contact(self,rock_id,contact_id):
		if contact_id == -1:
			self.contacts[rock_id]=self.contacts[rock_id][:contact_id]
		else:
			self.contacts[rock_id]=np.append(self.contacts[rock_id][:contact_id] , self.contacts[rock_id][contact_id+1:])
	
	def get_contacts_field(self,rock_id,field):
		return self.contacts[rock_id][:,self.contacts_slices[field]]
	def get_contacts_pos(self,rock_id):
		return self.get_contacts_field(rock_id,"rock_pos")
	def get_contacts_vels(self,rock_id):
		return self.get_contacts_field(rock_id,"rock_output_vels")
	def get_contacts_angVels(self,rock_id):
		if(self._ndim==2):
			return self.get_contacts_field(rock_id,"rock_output_angVels")[:,0]
		else:
			return self.get_contacts_field(rock_id,"rock_output_angVels")
	def get_contacts_types(self,rock_id):
		return self.get_contacts_field(rock_id,"types")[:,0]
	def get_contacts_normals(self,rock_id):
		return self.get_contacts_field(rock_id,"normals")

	def write_to_h5(self,filename):
		with h5py.File(filename, "w") as f:
			
			#GENERAL STRUCTURE:
			# hdf5-file/
			# ├── rocks/
			# |   ├── start_data
			# |   ├── contacts/
			# |   |   ├── 0 (contacts of rock n°0)
			# |   |   ├── ..
			# |   |   └── nb_rocks
			# └── checkpoints/
			#     ├── x_0 (x position of checkpoint 0)
			#     ├── x_1
			#     └── x_N

			f.create_group("rocks")
			f.create_group("rocks/contacts")
			f.create_group("checkpoints")
			
			#ROCKS START DATA:
			#prepare data:
			
			if self._ndim==2:
				inertias = self.inertias.reshape([self._s.nb_rocks,1])
			else :
				inertias = self.inertias.reshape([self._s.nb_rocks,9])
			start_data=np.concatenate((
				self.volumes.reshape([self._s.nb_rocks,1]),
				self.densities.reshape([self._s.nb_rocks,1]),
				inertias
			),axis=1).astype(float_type)
			#write to file:
			f.create_dataset("rocks/start_data",data=start_data)
			#add attributes that embed the links between start_data columns and data types
			f["rocks/start_data"].attrs["volumes"]=0
			f["rocks/start_data"].attrs["densities"]=1
			if(self._ndim==2):
				f["rocks/start_data"].attrs["inertias"]=2
			else:
				f["rocks/start_data"].attrs["inertias"]=[2,10]
			
			#ROCKS CONTACTS DATA:
			#loop on "rocks"
			for i in range(len(self.contacts)):
				#write the contacts of the rock i
				f.create_dataset("rocks/contacts/"+str(i),data=self.contacts[i].astype(float_type))
			#add attributes that embed the links between contacts data columns and data types
			for k in self.contacts_slices.keys():
				f["rocks/contacts"].attrs[k]=[self.contacts_slices[k].start,self.contacts_slices[k].stop-1]
			
			#CHECKPOINTS:
			if(self._s.checkpoints is not None):
				for c in self.checkpoints:
					data_len=len(c.rocks_ids)
					if(data_len):
						if self._ndim == 2 :
							angVels=np.asarray(c.angVels).reshape([data_len,1])
							heights2D_or_pos3D=np.asarray(c.heights).reshape([data_len,1])
						else:
							angVels=np.asarray(c.angVels)
							heights2D_or_pos3D=np.asarray(c.pos)
						chckpt_data=np.concatenate((
							np.asarray(c.rocks_ids).reshape([data_len,1]),
							heights2D_or_pos3D,
							np.asarray(c.vels),
							angVels,
						),axis=1)
					else:
						chckpt_data=np.array([])
					if(self._ndim==2):
						chkP_name=str(c.x)
					else:
						chkP_name=str(self.checkpoints.index(c))
					f.create_dataset("checkpoints/"+chkP_name,data=chckpt_data.astype(float_type))
					f["checkpoints"].attrs["rocks_ids"]=[0,0]
					if(self._ndim==2):
						f["checkpoints"].attrs["heights"]=[1,1]
						f["checkpoints"].attrs["vels"]=[2,3]
						f["checkpoints"].attrs["angVels"]=[4,4]
					else:
						f["checkpoints"].attrs["pos"]=[1,3]
						f["checkpoints"].attrs["vels"]=[4,6]
						f["checkpoints"].attrs["angVels"]=[7,9]
				
